# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from typing import List, Tuple, Dict

from xpra.x11.bindings.display_source import get_display_name   # @UnresolvedImport
from xpra.log import Logger


from xpra.x11.bindings.core import call_context_check  # @UnresolvedImport
from xpra.x11.bindings.core cimport X11CoreBindingsInstance
from xpra.x11.bindings.xlib cimport (
    XImage, Display, Pixmap,
    XColor, Visual, VisualID, XVisualInfo, VisualIDMask,
    Status, Window, Drawable, Bool,
    XFree, XWindowAttributes,
    MSBFirst, LSBFirst, ZPixmap,
    DoRed, DoGreen, DoBlue, AllPlanes,
    XGetVisualInfo,
)
from libc.stdlib cimport free
from libc.string cimport memcpy
from libc.stdint cimport uint64_t, uintptr_t

xshmlog = Logger("x11", "bindings", "ximage", "xshm")
log = Logger("x11", "bindings", "ximage")
xshmdebug = Logger("x11", "bindings", "ximage", "xshm", "verbose")
ximagedebug = Logger("x11", "bindings", "ximage", "verbose")


cdef inline unsigned int roundup(unsigned int n, unsigned int m) noexcept:
    return (n + m - 1) & ~(m - 1)


cdef inline unsigned char BYTESPERPIXEL(unsigned int depth) noexcept:
    if depth>=24 and depth<=32:
        return 4
    elif depth==16:
        return 2
    elif depth==8:
        return 1
    #shouldn't happen!
    return roundup(depth, 8)//8


cdef inline unsigned int MIN(unsigned int a, unsigned int b) noexcept:
    if a<=b:
        return a
    return b


###################################
# Headers, python magic
###################################
cdef extern from "Python.h":
    object PyMemoryView_FromMemory(char *mem, Py_ssize_t size, int flags)
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS
    int PyBUF_READ
    int PyBUF_WRITE

cdef extern from "stdlib.h":
    int posix_memalign(void **memptr, size_t alignment, size_t size)

cdef extern from "sys/ipc.h":
    ctypedef struct key_t:
        pass
    key_t IPC_PRIVATE
    int IPC_CREAT
    int IPC_RMID

cdef extern from "sys/shm.h":
    int shmget(key_t __key, size_t __size, int __shmflg)
    void *shmat(int __shmid, const void *__shmaddr, int __shmflg)
    int shmdt (const void *__shmaddr)
    ctypedef struct shmid_ds:
        pass
    int shmctl(int shmid, int cmd, shmid_ds *buf)

cdef extern from "errno.h" nogil:
    int errno
    int EINVAL

ctypedef unsigned long CARD32
ctypedef unsigned short CARD16
ctypedef CARD32 Colormap
DEF XNone = 0

cdef extern from "X11/X.h":
    unsigned long NoSymbol

cdef extern from "X11/Xatom.h":
    int XA_RGB_DEFAULT_MAP
    int XA_RGB_BEST_MAP

cdef extern from "X11/Xlib.h":
    void XQueryColors(Display *display, Colormap colormap, XColor defs_in_out[], int ncolors)
    VisualID XVisualIDFromVisual(Visual *visual)

    Status XGetWindowAttributes(Display * display, Window w, XWindowAttributes * attributes)


    XImage *XGetImage(Display *display, Drawable d,
            int x, int y, unsigned int width, unsigned int  height,
            unsigned long plane_mask, int fmt)

    void XDestroyImage(XImage *ximage)

    Status XGetGeometry(Display *display, Drawable d, Window *root_return,
                        int *x_return, int *y_return, unsigned int  *width_return, unsigned int *height_return,
                        unsigned int *border_width_return, unsigned int *depth_return)

cdef extern from "X11/Xutil.h":
    pass

cdef extern from "X11/extensions/Xcomposite.h":
    Bool XCompositeQueryExtension(Display *, int * major, int * minor)
    Status XCompositeQueryVersion(Display *, int * major, int * minor)
    int XFreePixmap(Display *, Pixmap pixmap)
    Pixmap XCompositeNameWindowPixmap(Display *xdisplay, Window xwindow)

cdef extern from "X11/extensions/XShm.h":
    unsigned int ShmCompletion
    ctypedef struct ShmSeg:
        pass
    ctypedef struct XShmSegmentInfo:
        ShmSeg shmseg   # resource id
        int shmid       # kernel id
        char *shmaddr   # address in client
        Bool readOnly   # how the server should attach it

    Bool XShmQueryExtension(Display *display)
    Bool XShmQueryVersion(Display *display, int *major, int *minor, Bool *pixmaps)

    Bool XShmAttach(Display *display, XShmSegmentInfo *shminfo)
    Bool XShmDetach(Display *display, XShmSegmentInfo *shminfo)

    XImage *XShmCreateImage(Display *display, Visual *visual,
                            unsigned int depth, int fmt, char *data,
                            XShmSegmentInfo *shminfo,
                            unsigned int width, unsigned int height)

    Bool XShmGetImage(Display *display, Drawable d, XImage *image,
                      int x, int y,
                      unsigned long plane_mask)

    int XShmGetEventBase(Display *display)


SBFirst: Dict[int, str] = {
    MSBFirst : "MSBFirst",
    LSBFirst : "LSBFirst"
}

RLE8    = "RLE8"
RGB565  = "RGB565"
BGR565  = "BGR565"
XRGB    = "XRGB"
BGRX    = "BGRX"
ARGB    = "ARGB"
BGRA    = "BGRA"
RGB     = "RGB"
RGBA    = "RGBA"
RGBX    = "RGBX"
R210    = "R210"
r210    = "r210"

RGB_FORMATS = [XRGB, BGRX, ARGB, BGRA, RGB, RGBA, RGBX, R210, r210, RGB565, BGR565, RLE8]


cdef int ximage_counter = 0

cdef class XImageWrapper:
    """
        Presents X11 image pixels as in ImageWrapper
    """

    cdef XImage *image
    cdef unsigned int x
    cdef unsigned int y
    cdef unsigned int target_x
    cdef unsigned int target_y
    cdef unsigned int width
    cdef unsigned int height
    cdef unsigned int depth
    cdef unsigned int rowstride
    cdef unsigned int planes
    cdef unsigned int bytesperpixel
    cdef unsigned char thread_safe
    cdef unsigned char sub
    cdef object pixel_format
    cdef void *pixels
    cdef object del_callback
    cdef uint64_t timestamp
    cdef object palette
    cdef unsigned char full_range

    def __cinit__(self, unsigned int x, unsigned int y, unsigned int width, unsigned int height,
                        uintptr_t pixels=0, pixel_format="", unsigned int depth=24, unsigned int rowstride=0,
                        int planes=0, unsigned int bytesperpixel=4, thread_safe=False, sub=False,
                        palette=None, full_range=True):
        self.image = NULL
        self.pixels = NULL
        self.x = x
        self.y = y
        self.target_x = x
        self.target_y = y
        self.width = width
        self.height = height
        self.depth = depth
        self.bytesperpixel = bytesperpixel
        self.pixel_format = pixel_format
        self.rowstride = rowstride
        self.planes = planes
        self.thread_safe = thread_safe
        self.sub = sub
        self.pixels = <void *> pixels
        self.timestamp = int(monotonic()*1000)
        self.palette = palette
        self.full_range = int(full_range)

    cdef void set_image(self, XImage* image):
        assert not self.sub
        assert image!=NULL
        global ximage_counter
        ximage_counter += 1
        self.thread_safe = 0
        self.image = image
        self.rowstride = image.bytes_per_line
        self.depth = image.depth
        if self.depth==24:
            self.bytesperpixel = 4
            if image.byte_order==MSBFirst:
                self.pixel_format = XRGB
            else:
                self.pixel_format = BGRX
        elif self.depth==16:
            self.bytesperpixel = 2
            if image.byte_order==MSBFirst:
                self.pixel_format = RGB565
            else:
                self.pixel_format = BGR565
        elif self.depth==8:
            self.bytesperpixel = 1
            self.pixel_format = RLE8
        elif self.depth==32:
            self.bytesperpixel = 4
            if image.byte_order==MSBFirst:
                self.pixel_format = ARGB
            else:
                self.pixel_format = BGRA
        elif self.depth==30:
            #log.warn("30bpp, byte_order=%i, bitmap_bit_order=%i, bits_per_pixel=%s", image.byte_order, image.bitmap_bit_order, image.bits_per_pixel)
            #log.warn("pad=%i, unit=%i, format=%i", image.bitmap_pad, image.bitmap_unit, image.format)
            #log.warn("masks: red=%#x, green=%#x, blue=%#x", image.red_mask, image.green_mask, image.blue_mask)
            self.bytesperpixel = 4
            if image.byte_order==MSBFirst:
                self.pixel_format = R210
            else:
                self.pixel_format = r210
        else:
            raise ValueError(f"invalid image depth: {self.depth} bpp")

    def __repr__(self):
        return "XImageWrapper(%s: %i, %i, %i, %i)" % (self.pixel_format, self.x, self.y, self.width, self.height)

    def get_geometry(self) -> Tuple[int, int, int, int, int]:
        return self.x, self.y, self.width, self.height, self.depth

    def get_target_x(self) -> int:
        return self.target_x

    def get_target_y(self) -> int:
        return self.target_y

    def set_target_x(self, unsigned int target_x):
        self.target_x = target_x

    def set_target_y(self, unsigned int target_y):
        self.target_y = target_y

    def get_x(self) -> int:
        return self.x

    def get_y(self) -> int:
        return self.y

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_rowstride(self) -> int:
        return self.rowstride

    def get_palette(self):
        return self.palette

    def get_full_range(self) -> bool:
        return bool(self.full_range)

    def get_planes(self) -> int:
        return self.planes

    def get_depth(self) -> int:
        return self.depth

    def get_bytesperpixel(self):
        return self.bytesperpixel

    def get_size(self) -> int:
        return self.rowstride * self.height

    def get_pixel_format(self):
        return self.pixel_format

    def get_pixels(self):
        cdef void *pix_ptr = self.get_pixels_ptr()
        if pix_ptr==NULL:
            return None
        cdef int flags = PyBUF_READ
        if self.pixels!=NULL:
            flags = PyBUF_WRITE
        return PyMemoryView_FromMemory(<char *> pix_ptr, self.get_size(), PyBUF_READ)

    def get_sub_image(self, unsigned int x, unsigned int y, unsigned int w, unsigned int h):
        if w<=0 or h<=0:
            raise ValueError(f"invalid sub-image size: {w}x{h}")
        if x+w>self.width:
            raise ValueError(f"invalid sub-image width: {x}+{w} greater than image width {self.width}")
        if y+h>self.height:
            raise ValueError(f"invalid sub-image height: {y}+{h} greater than image height {self.height}")
        cdef void *src = self.get_pixels_ptr()
        if src==NULL:
            raise ValueError("source image does not have any pixels!")
        cdef unsigned char Bpp = BYTESPERPIXEL(self.depth)
        cdef uintptr_t sub_ptr = (<uintptr_t> src) + x*Bpp + y*self.rowstride
        image = XImageWrapper(self.x+x, self.y+y, w, h, sub_ptr, self.pixel_format,
                             self.depth, self.rowstride, self.planes, self.bytesperpixel, True, True,
                             self.palette, self.full_range)
        image.set_target_x(self.target_x+x)
        image.set_target_y(self.target_y+y)
        return image

    cdef void* get_pixels_ptr(self):
        if self.pixels!=NULL:
            return self.pixels
        cdef XImage *image = self.image
        if image==NULL:
            log.warn("get_pixels_ptr: image is NULL!")
            return NULL
        if image.data is NULL:
            log.warn("get_pixels_ptr: image.data is NULL!")
        return image.data

    def get_gpu_buffer(self):
        return None

    def has_pixels(self) -> bool:
        if self.pixels!=NULL:
            return True
        cdef XImage *image = self.image
        if image==NULL:
            return False
        return image.data!=NULL

    def is_thread_safe(self) -> bool:
        return self.thread_safe

    def get_timestamp(self) -> int:
        """ time in millis """
        return self.timestamp

    def set_palette(self, palette) -> None:
        self.palette = palette

    def set_full_range(self, full_range: bool) -> None:
        self.full_range = int(full_range)

    def set_timestamp(self, timestamp) -> None:
        self.timestamp = timestamp

    def set_rowstride(self, unsigned int rowstride) -> None:
        self.rowstride = rowstride

    def set_pixel_format(self, pixel_format) -> None:
        assert pixel_format is not None and pixel_format in RGB_FORMATS, "invalid pixel format: %s" % pixel_format
        self.pixel_format = pixel_format

    def set_pixels(self, pixels) -> None:
        """ overrides the context of the image with the given pixels buffer """
        if self.pixels!=NULL:
            if not self.sub:
                free(self.pixels)
            self.pixels = NULL

        cdef Py_buffer py_buf
        if PyObject_GetBuffer(pixels, &py_buf, PyBUF_ANY_CONTIGUOUS):
            raise ValueError(f"failed to read pixel data from {type(pixels)}")

        #Note: we can't free the XImage, because it may
        #still be used somewhere else (see XShmWrapper)
        if posix_memalign(<void **> &self.pixels, 64, py_buf.len):
            PyBuffer_Release(&py_buf)
            raise RuntimeError("posix_memalign failed!")
        assert self.pixels!=NULL
        #from now on, we own the buffer,
        #so we're no longer a direct sub-image,
        #and we must free the buffer later:
        self.sub = False
        if self.image==NULL:
            self.thread_safe = 1
            #we can now mark this object as thread safe
            #if we have already freed the XImage
            #which needs to be freed from the UI thread
            #but our new buffer is just a malloc buffer,
            #which is safe from any thread
        memcpy(self.pixels, py_buf.buf, py_buf.len)
        PyBuffer_Release(&py_buf)

    def free(self) -> None:
        ximagedebug("%s.free()", self)
        self.free_image()
        self.free_pixels()

    cdef void free_image(self):
        ximagedebug("%s.free_image() image=%#x", self, <uintptr_t> self.image)
        if self.image!=NULL:
            call_context_check("XImageWrapper.free_image")
            XDestroyImage(self.image)
            self.image = NULL
            global ximage_counter
            ximage_counter -= 1

    cdef void free_pixels(self):
        ximagedebug("%s.free_pixels() pixels=%#x", self, <uintptr_t> self.pixels)
        if self.pixels!=NULL:
            if not self.sub:
                free(self.pixels)
            self.pixels = NULL

    def freeze(self) -> bool:
        #we don't need to do anything here because the non-XShm version
        #already uses a copy of the pixels
        return False

    def may_restride(self) -> bool:
        #if not given a newstride, assume it is optional and check if it is worth doing at all:
        if self.rowstride<=8 or self.height<=2:
            return False                                    #not worth it
        #use a reasonable stride: rounded up to 4
        cdef unsigned int newstride = roundup(self.width*len(self.pixel_format), 4)
        if newstride>=self.rowstride:
            return False                                    #not worth it
        cdef unsigned int newsize = newstride*self.height   #desirable size we could have
        cdef unsigned int size = self.rowstride*self.height
        if size-newsize<1024 or newsize*110/100>size:
            log("may_restride() not enough savings with new stride=%i vs %i: new size=%i vs %i", newstride, self.rowstride, newsize, size)
            return False
        return self.restride(newstride)

    def restride(self, const unsigned int rowstride) -> bool:
        #NOTE: this must be called from the UI thread!
        #start = monotonic()
        cdef unsigned int newsize = rowstride*self.height                #desirable size we could have
        cdef unsigned int size = self.rowstride*self.height
        #is it worth re-striding to save space:
        #(save at least 1KB and 10%)
        #Note: we could also change the pixel format whilst we're at it
        # and convert BGRX to RGB for example (assuming RGB is also supported by the client)
        cdef void *img_buf = self.get_pixels_ptr()
        assert img_buf!=NULL, "this image wrapper is empty!"
        cdef void *new_buf
        if posix_memalign(<void **> &new_buf, 64, (newsize+rowstride)):
            raise RuntimeError("posix_memalign failed!")
        cdef void *to = new_buf
        cdef unsigned int oldstride = self.rowstride                     #using a local variable is faster
        #Note: we don't zero the buffer,
        #so if the newstride is bigger than oldstride, you get garbage..
        cdef unsigned int cpy_size
        if oldstride==rowstride:
            memcpy(to, img_buf, size)
        else:
            cpy_size = MIN(rowstride, oldstride)
            for _ in range(self.height):
                memcpy(to, img_buf, cpy_size)
                to += rowstride
                img_buf += oldstride
        #we can now free the pixels buffer if present
        #(but not the ximage - this is not running in the UI thread!)
        self.free_pixels()
        #set the new attributes:
        self.rowstride = rowstride
        self.pixels = <char *> new_buf
        #without any X11 image to free, this is now thread safe:
        if self.image==NULL:
            self.thread_safe = 1
        #log("restride(%s) %s pixels re-stride saving %i%% from %s (%s bytes) to %s (%s bytes) took %.1fms",
        #    rowstride, self.pixel_format, 100-100*newsize/size, oldstride, size, rowstride, newsize, (monotonic()-start)*1000)
        return True


cdef class XShmWrapper:
    cdef Display *display
    cdef Visual *visual
    cdef Window window
    cdef unsigned int width
    cdef unsigned int height
    cdef unsigned int depth
    cdef XShmSegmentInfo shminfo
    cdef XImage *image
    cdef unsigned int ref_count
    cdef Bool got_image
    cdef Bool closed

    cdef void init(self, Display *display, Window xwindow, Visual *visual, unsigned int width, unsigned int height, unsigned int depth):
        self.display = display
        self.window = xwindow
        self.visual = visual
        self.width = width
        self.height = height
        self.depth = depth

    def __repr__(self):
        return "XShmWrapper(%#x - %ix%i)" % (self.window, self.width, self.height)

    def setup(self) -> Tuple[bool, bool, bool]:
        #returns:
        # (init_ok, may_retry_this_window, XShm_global_failure)
        self.ref_count = 0
        self.closed = False
        self.shminfo.shmaddr = <char *> -1

        self.image = XShmCreateImage(self.display, self.visual, self.depth,
                          ZPixmap, NULL, &self.shminfo,
                          self.width, self.height)
        xshmdebug("XShmWrapper.setup() XShmCreateImage(%ix%i-%i) %s", self.width, self.height, self.depth, self.image!=NULL)
        if self.image==NULL:
            xshmlog.error("XShmWrapper.setup() XShmCreateImage(%ix%i-%i) failed!", self.width, self.height, self.depth)
            self.cleanup()
            #if we cannot create an XShm XImage, we may try again
            #(it could be dimensions are too big?)
            return False, True, False
        # Get the shared memory:
        # (include an extra line to ensure we can read rowstride at a time,
        #  even on the last line, without reading past the end of the buffer)
        cdef size_t size = self.image.bytes_per_line * (self.image.height + 1)
        self.shminfo.shmid = shmget(IPC_PRIVATE, size, IPC_CREAT | 0o777)
        xshmdebug("XShmWrapper.setup() shmget(PRIVATE, %i bytes, %#x) shmid=%#x", size, IPC_CREAT | 0777, self.shminfo.shmid)
        if self.shminfo.shmid < 0:
            xshmlog.error("XShmWrapper.setup() shmget(PRIVATE, %i bytes, %#x) failed, bytes_per_line=%i, width=%i, height=%i", size, IPC_CREAT | 0777, self.image.bytes_per_line, self.width, self.height)
            self.cleanup()
            #only try again if we get EINVAL,
            #the other error codes probably mean this is never going to work..
            return False, errno==EINVAL, errno!=EINVAL
        # Attach:
        self.image.data = <char *> shmat(self.shminfo.shmid, NULL, 0)
        self.shminfo.shmaddr = self.image.data
        xshmdebug("XShmWrapper.setup() shmat(%s, NULL, 0) %s", self.shminfo.shmid, self.shminfo.shmaddr != <char *> -1)
        if self.shminfo.shmaddr == <char *> -1:
            xshmlog.error("XShmWrapper.setup() shmat(%s, NULL, 0) failed!", self.shminfo.shmid)
            self.cleanup()
            #we may try again with this window, or any other window:
            #(as this really shouldn't happen at all)
            return False, True, False

        # set as read/write, and attach to the display:
        self.shminfo.readOnly = False
        cdef Bool a = XShmAttach(self.display, &self.shminfo)
        xshmdebug("XShmWrapper.setup() XShmAttach(..) %s", bool(a))
        if not a:
            xshmlog.error("XShmWrapper.setup() XShmAttach(..) failed!")
            self.cleanup()
            #we may try again with this window, or any other window:
            #(as this really shouldn't happen at all)
            return False, True, False
        return True, True, False

    def get_size(self) -> Tuple[int, int]:
        return self.width, self.height

    def get_image(self, Drawable drawable, int x, int y, int w, int h):
        assert self.image!=NULL, "cannot retrieve image wrapper: XImage is NULL!"
        if self.closed:
            return None
        cdef int maxw = self.width
        cdef int maxh = self.height
        if w+x<=0 or h+y<=0:
            xshmlog("XShmWrapper.get_image%s invalid position + dimensions", (drawable, x, y, w, h))
            return None
        if x<0:
            w += x
            x = 0
        elif x>=maxw:
            xshmlog("XShmWrapper.get_image%s x-value is outside image dimensions %ix%i", (drawable, x, y, w, h), self.width, self.height)
            return None
        if y<0:
            h += y
            y = 0
        elif y>=maxh:
            xshmlog("XShmWrapper.get_image%s y-value is outside image dimensions %ix%i", (drawable, x, y, w, h), self.width, self.height)
            return None
        #clamp size to image size:
        if x+w>maxw:
            w = maxw-x
            assert w>0
        if y+h>maxh:
            h = maxh-y
            assert h>0
        if not self.got_image:
            if not XShmGetImage(self.display, drawable, self.image, 0, 0, 0xFFFFFFFF):
                xshmlog("XShmWrapper.get_image(%#x, %i, %i, %i, %i) XShmGetImage failed!", drawable, x, y, w, h)
                return None
            self.got_image = True
        self.ref_count += 1
        cdef XShmImageWrapper imageWrapper = XShmImageWrapper(x, y, w, h)
        imageWrapper.set_image(self.image)
        imageWrapper.set_free_callback(self.free_image_callback)
        if self.depth==8:
            imageWrapper.set_palette(self.read_palette())
        xshmdebug("XShmWrapper.get_image(%#x, %i, %i, %i, %i)=%s (ref_count=%i)", drawable, x, y, w, h, imageWrapper, self.ref_count)
        return imageWrapper

    def read_palette(self) -> List[Tuple[int, int, int]]:
        cdef XWindowAttributes attrs
        cdef XVisualInfo vinfo_template
        cdef int count
        if not XGetWindowAttributes(self.display, self.window, &attrs):
            return None
        cdef Colormap colormap = attrs.colormap
        cdef VisualID visualid = XVisualIDFromVisual(attrs.visual)
        vinfo_template.visualid = visualid
        cdef XVisualInfo *vinfo = XGetVisualInfo(self.display, VisualIDMask, &vinfo_template, &count)
        if count!=1 or vinfo==NULL:
            log.error("Error: visual %i not found, count=%i, vinfo=%#x", visualid, count, <uintptr_t> vinfo)
            if vinfo:
                XFree(vinfo)
            return None
        log("visual: depth=%i, red mask=%#10x, green mask=%#10x, blue mask=%#10x, size=%i, bits per rgb=%i", vinfo.depth, vinfo.red_mask, vinfo.green_mask, vinfo.blue_mask, vinfo.colormap_size, vinfo.bits_per_rgb)
        cdef unsigned int size = vinfo.colormap_size
        XFree(vinfo)
        if size>256:
            log.error("invalid colormap size: %i", size)
            return None
        cdef XColor[256] colors
        cdef unsigned int i
        for i in range(size):
            colors[i].flags = DoRed | DoGreen | DoBlue
            colors[i].pixel = i
        for i in range(size, 256):
            colors[i].flags = 0
            colors[i].pixel = 0
        XQueryColors(self.display, colormap, colors, size)
        palette = [(colors[i].red, colors[i].green, colors[i].blue) for i in range(256)]
        return palette

    def discard(self) -> None:
        #force next get_image call to get a new image from the server
        self.got_image = False

    def __dealloc__(self):
        xshmdebug("XShmWrapper.__dealloc__() ref_count=%i", self.ref_count)
        self.cleanup()

    def cleanup(self) -> None:
        #ok, we want to free resources... problem is,
        #we may have handed out some XShmImageWrappers
        #and they will point to our Image XShm area.
        #so we have to wait until *they* are freed,
        #and rely on them telling us via the free_image_callback.
        xshmdebug("XShmWrapper.cleanup() ref_count=%i", self.ref_count)
        self.closed = True
        if self.ref_count==0:
            self.free()

    def free_image_callback(self) -> None:
        self.ref_count -= 1
        xshmdebug("XShmWrapper.free_image_callback() closed=%s, new ref_count=%i", self.closed, self.ref_count)
        if self.closed and self.ref_count==0:
            self.free()

    cdef void free(self):
        assert self.ref_count==0, "XShmWrapper %s cannot be freed: still has a ref count of %i" % (self, self.ref_count)
        assert self.closed, "XShmWrapper %s cannot be freed: it is not closed yet" % self
        has_shm = self.shminfo.shmaddr!=<char *> -1
        xshmdebug("XShmWrapper.free() has_shm=%s, image=%#x, shmid=%#x", has_shm, <uintptr_t> self.image, self.shminfo.shmid)
        if has_shm:
            XShmDetach(self.display, &self.shminfo)
        has_image = self.image!=NULL
        if has_image:
            XDestroyImage(self.image)
            self.image = NULL
        if has_shm:
            shmctl(self.shminfo.shmid, IPC_RMID, NULL)
            shmdt(self.shminfo.shmaddr)
            self.shminfo.shmaddr = <char *> -1
            self.shminfo.shmid = -1
        if has_shm or has_image:
            call_context_check("XShmWrapper.free")


cdef class XShmImageWrapper(XImageWrapper):

    cdef object free_callback

    def __init__(self, *args):
        self.free_callback = None

    def __repr__(self):
        return "XShmImageWrapper(%s: %s, %s, %s, %s)" % (self.pixel_format, self.x, self.y, self.width, self.height)

    cdef void *get_pixels_ptr(self):
        if self.pixels!=NULL:
            xshmdebug("XShmImageWrapper.get_pixels_ptr()=%#x (pixels) %s", <uintptr_t> self.pixels, self)
            return self.pixels
        cdef XImage *image = self.image             #
        if image==NULL:
            xshmdebug("XShm get_pixels_ptr XImage is NULL")
            return NULL
        assert self.height>0
        #calculate offset (assuming 4 bytes "pixelstride"):
        cdef unsigned char Bpp = BYTESPERPIXEL(self.depth)
        cdef void *ptr = image.data + (self.y * self.rowstride) + (Bpp * self.x)
        xshmdebug("XShmImageWrapper.get_pixels_ptr()=%#x %s", <uintptr_t> ptr, self)
        return ptr

    def freeze(self) -> bool:
        #we just force a restride, which will allocate a new pixel buffer:
        cdef unsigned int newstride = roundup(self.width*len(self.pixel_format), 4)
        self.timestamp = int(monotonic() * 1000)
        return self.restride(newstride)

    def free(self) -> None:
        #ensure we never try to XDestroyImage:
        self.image = NULL
        self.free_pixels()
        cb = self.free_callback
        if cb:
            self.free_callback = None
            cb()
        xshmdebug("XShmImageWrapper.free() done for %s, callback fired=%s", self, bool(cb))

    cdef void set_free_callback(self, object callback):
        self.free_callback = callback


cdef int xpixmap_counter = 0


cdef class PixmapWrapper:
    cdef Display *display
    cdef Pixmap pixmap
    cdef unsigned int width
    cdef unsigned int height

    "Reference count an X Pixmap that needs explicit cleanup."
    cdef void init(self, Display *display, Pixmap pixmap, unsigned int width, unsigned int height):
        self.display = display
        self.pixmap = pixmap
        self.width = width
        self.height = height
        global xpixmap_counter
        xpixmap_counter += 1
        ximagedebug("%s xpixmap counter: %i", self, xpixmap_counter)

    def __repr__(self):
        return "PixmapWrapper(%#x, %i, %i)" % (self.pixmap, self.width, self.height)

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_pixmap(self) -> Pixmap:
        return self.pixmap

    def get_image(self, unsigned int x, unsigned int y, unsigned int width, unsigned int height):
        if not self.pixmap:
            log.warn("%s.get_image%s", self, (x, y, width, height))
            return  None
        ximagedebug("%s.get_image%s width=%i, height=%i", self, (x, y, width, height), self.width, self.height)
        if x>=self.width or y>=self.height:
            log("%s.get_image%s position outside image dimensions %ix%i", self, (x, y, width, height), self.width, self.height)
            return None
        #clamp size to image size:
        if x+width>self.width:
            width = self.width-x
        if y+height>self.height:
            height = self.height-y
        return get_image(self.display, self.pixmap, x, y, width, height)

    def __dealloc__(self):
        self.do_cleanup()

    cdef void do_cleanup(self):
        if self.pixmap!=0:
            XFreePixmap(self.display, self.pixmap)
            self.pixmap = 0
            global xpixmap_counter
            xpixmap_counter -= 1

    def cleanup(self) -> None:
        ximagedebug("%s.cleanup()", self)
        self.do_cleanup()



cdef XImageWrapper get_image(Display * display, Drawable drawable, unsigned int x, unsigned int y, unsigned int width, unsigned int height):
    cdef XImage* ximage = XGetImage(display, drawable, x, y, width, height, AllPlanes, ZPixmap)
    #log.info("get_pixels(..) ximage==NULL : %s", ximage==NULL)
    if ximage==NULL:
        log("get_image(..) failed to get XImage for X11 drawable %#x", drawable)
        return None
    xi = XImageWrapper(x, y, width, height)
    xi.set_image(ximage)
    global ximage_counter
    ximagedebug("%s ximage counter: %i", xi, ximage_counter)
    return xi


cdef object xcomposite_name_window_pixmap(Display * xdisplay, Window xwindow):
    cdef Window root_window
    cdef int x, y
    cdef unsigned int width, height, border, depth
    cdef Pixmap xpixmap = XCompositeNameWindowPixmap(xdisplay, xwindow)
    if xpixmap==XNone:
        return None
    cdef Status status = XGetGeometry(xdisplay, xpixmap, &root_window,
                        &x, &y, &width, &height, &border, &depth)
    if status==0:
        log("failed to get pixmap dimensions for %#x", xpixmap)
        XFreePixmap(xdisplay, xpixmap)
        return None
    cdef PixmapWrapper pw = PixmapWrapper()
    pw.init(xdisplay, xpixmap, width, height)
    return pw


cdef object window_pixmap_wrapper(Display *xdisplay, Window xwindow):
    cdef Window root_window
    cdef int x, y
    cdef unsigned width, height, border, depth
    status = XGetGeometry(xdisplay, xwindow, &root_window,
                        &x, &y, &width, &height, &border, &depth)
    if status==0:
        log("failed to get window dimensions for %#x", xwindow)
        return None
    cdef PixmapWrapper pw = PixmapWrapper()
    pw.init(xdisplay, xwindow, width, height)
    return pw


cdef class XImageBindingsInstance(X11CoreBindingsInstance):
    cdef int has_xshm

    def __cinit__(self):
        self.has_xshm = XShmQueryExtension(self.display)
        dn = get_display_name()
        xshmlog("XShmQueryExtension()=%s on display {dn!r}", bool(self.has_xshm))
        if not self.has_xshm:
            xshmlog.warn(f"Warning: no XShm support on display {dn!r}")

    def has_XShm(self) -> bool:
        return bool(self.has_xshm)

    def get_XShmWrapper(self, xwindow):
        self.context_check("get_XShmWrapper")
        cdef XWindowAttributes attrs
        if XGetWindowAttributes(self.display, xwindow, &attrs)==0:
            return None
        xshm = XShmWrapper()
        xshm.init(self.display, xwindow, attrs.visual, attrs.width, attrs.height, attrs.depth)
        return xshm

    def get_ximage(self, drawable, x, y, width, height):
        self.context_check("get_ximage")
        return get_image(self.display, drawable, x, y, width, height)

    def get_xcomposite_pixmap(self, xwindow):
        self.context_check("get_xcomposite_pixmap")
        return xcomposite_name_window_pixmap(self.display, xwindow)

    def get_xwindow_pixmap_wrapper(self, xwindow):
        self.context_check("get_xwindow_pixmap_wrapper")
        return window_pixmap_wrapper(self.display, xwindow)


cdef XImageBindingsInstance singleton = None


def XImageBindings() -> XImageBindingsInstance:
    global singleton
    if singleton is None:
        singleton = XImageBindingsInstance()
    return singleton




# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

def verify_error(msg, *args) -> None:
    from xpra.log import Logger
    log = Logger("network")
    log.error(msg, *args)


def verify_packet(packet) -> bool:
    """ look for None values which may have caused the packet to fail encoding """
    if not isinstance(packet, (list, tuple)):
        return False
    if not packet:
        raise ValueError(f"invalid packet: {packet} ({type(packet)})")
    tree = [f"{packet[0]!r} packet"]
    return do_verify_packet(tree, packet)


def do_verify_packet(tree, packet) -> bool:
    def err(msg):
        verify_error("%s in %s", msg, "->".join(tree))

    def new_tree(append):
        nt = tree[:]
        nt.append(append)
        return nt

    if packet is None:
        err("None value")
        return False
    r = True
    if isinstance(packet, (list, tuple)):
        for i, x in enumerate(packet):
            if not do_verify_packet(new_tree(f"[{i}]"), x):
                r = False
    elif isinstance(packet, dict):
        for k, v in packet.items():
            if not do_verify_packet(new_tree(f"key for value={v!r}"), k):
                r = False
            if not do_verify_packet(new_tree(f"value for key={k!r}"), v):
                r = False
    elif isinstance(packet, (int, bool, str, bytes, memoryview)):
        "OK - supported"
    else:
        err(f"unsupported type: {type(packet)}")
        r = False
    return r



# colored stream handler for python logging framework (use the ColorStreamHandler class).
# based on:
# http://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output/1336640#1336640
#
# Copyright (c) 2014 Markus Pointner
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
import logging


class _AnsiColorStreamHandler(logging.StreamHandler):
    DEFAULT = '\x1b[0m'
    RED = '\x1b[31m'
    GREEN = '\x1b[32m'
    YELLOW = '\x1b[33m'
    CYAN = '\x1b[36m'
    BLUE = '\x1b[34m'
    MAGENTA = '\x1b[35m'
    NONE = ''

    CRITICAL = MAGENTA
    ERROR = RED
    WARNING = YELLOW
    INFO = NONE
    DEBUG = GREEN

    @classmethod
    def _get_color(cls, level):
        if level >= logging.CRITICAL:
            return cls.CRITICAL
        if level >= logging.ERROR:
            return cls.ERROR
        if level >= logging.WARNING:
            return cls.WARNING
        if level >= logging.INFO:
            return cls.INFO
        if level >= logging.DEBUG:
            return cls.DEBUG
        return cls.DEFAULT

    def format(self, record):
        try:
            text = super().format(record)
        except Exception:
            text = str(record)
        color = self._get_color(record.levelno)
        return color + text + self.DEFAULT


class _WinColorStreamHandler(logging.StreamHandler):
    # wincon.h
    FOREGROUND_BLACK = 0x0000
    FOREGROUND_BLUE = 0x0001
    FOREGROUND_GREEN = 0x0002
    FOREGROUND_CYAN = 0x0003
    FOREGROUND_RED = 0x0004
    FOREGROUND_MAGENTA = 0x0005
    FOREGROUND_YELLOW = 0x0006
    FOREGROUND_GREY = 0x0007
    FOREGROUND_INTENSITY = 0x0008  # foreground color is intensified.
    FOREGROUND_WHITE = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED

    BACKGROUND_BLACK = 0x0000
    BACKGROUND_BLUE = 0x0010
    BACKGROUND_GREEN = 0x0020
    BACKGROUND_CYAN = 0x0030
    BACKGROUND_RED = 0x0040
    BACKGROUND_MAGENTA = 0x0050
    BACKGROUND_YELLOW = 0x0060
    BACKGROUND_GREY = 0x0070
    BACKGROUND_INTENSITY = 0x0080  # background color is intensified.

    DEFAULT = FOREGROUND_GREEN
    CRITICAL = BACKGROUND_YELLOW | FOREGROUND_MAGENTA | FOREGROUND_INTENSITY | BACKGROUND_INTENSITY
    ERROR = FOREGROUND_RED | FOREGROUND_INTENSITY
    WARNING = FOREGROUND_YELLOW | FOREGROUND_INTENSITY
    INFO = FOREGROUND_WHITE
    DEBUG = FOREGROUND_GREEN

    @classmethod
    def _get_color(cls, level):
        if level >= logging.CRITICAL:
            return cls.CRITICAL
        if level >= logging.ERROR:
            return cls.ERROR
        if level >= logging.WARNING:
            return cls.WARNING
        if level >= logging.INFO:
            return cls.INFO
        if level >= logging.DEBUG:
            return cls.DEBUG
        return cls.DEFAULT

    def _set_color(self, code):
        import ctypes
        ctypes.windll.kernel32.SetConsoleTextAttribute(self._outhdl, code)

    def __init__(self, stream=None):
        super().__init__(stream)
        # get file handle for the stream
        import ctypes.util
        crtname = ctypes.util.find_msvcrt()  # @UndefinedVariable
        crtlib = ctypes.cdll.LoadLibrary(crtname)
        self._outhdl = crtlib._get_osfhandle(stream.fileno())  # pylint: disable=protected-access

    def emit(self, record):
        color = self._get_color(record.levelno)
        self._set_color(color)
        super().emit(record)
        self._set_color(self.FOREGROUND_WHITE)


# select ColorStreamHandler based on platform
if sys.platform.startswith("win") and not os.environ.get("MSYSCON"):
    ColorStreamHandler = _WinColorStreamHandler
else:
    ColorStreamHandler = _AnsiColorStreamHandler



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.net.websockets.header import encode_hybi_header, decode_hybi
from xpra.log import Logger

log = Logger("network")


class WebsocketHeaderTest(unittest.TestCase):

    def test_round_trip(self):
        def rt(opcode, payload, has_mask=False, fin=True):
            h = encode_hybi_header(opcode, len(payload), has_mask, fin)
            packet = h+payload
            ropcode, rpayload, rlen, rfin = decode_hybi(packet)
            assert opcode==ropcode
            assert rpayload==payload
            assert rlen>len(payload)
            assert fin==rfin
        for l in (0, 10, 125, 126, 65535, 65536):
            rt(0, b"\0"*l)


    def test_invalid_decode(self):
        for l in (0, 10, 125, 126, 65535, 65536):
            for has_mask in (True, False):
                payload = b"\0"*l
                h = encode_hybi_header(0, len(payload), has_mask, True)
                if has_mask:
                    packet = h+b"9"*4+payload
                else:
                    packet = h+payload
                v = decode_hybi(packet)
                assert v
                v = decode_hybi(packet[:-1])
                assert v is None
                v = decode_hybi(packet[:int(has_mask)*4])
                assert v is None, "got %s" % (v,)
                if l>0:
                    for i in range(5):
                        v = decode_hybi(packet[:int(has_mask)*4+i])
                        assert v is None, "got %s" % (v,)

def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import io
from typing import Any
from contextlib import redirect_stdout, redirect_stderr

from xpra.util.objects import typedict
from xpra.scripts.config import TRUE_OPTIONS
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.log import Logger

log = Logger("exec")


class ShellMixin(StubSourceMixin):

    PREFIX = "shell"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        return caps.boolget(ShellMixin.PREFIX, False)

    def __init__(self, *_args):
        self._server = None
        self.shell_enabled = False
        self.saved_logging_handler = None
        self.log_records = []
        self.log_thread = None

    def init_from(self, protocol, server) -> None:
        self._server = server
        try:
            options = protocol._conn.options
            shell = options.get("shell", "")
            self.shell_enabled = shell.lower() in TRUE_OPTIONS
        except AttributeError:
            options = {}
            self.shell_enabled = False
        log("init_from(%s, %s) shell_enabled(%s)=%s", protocol, server, options, self.shell_enabled)

    def get_caps(self) -> dict[str, Any]:
        return {ShellMixin.PREFIX: self.shell_enabled}

    def get_info(self) -> dict[str, Any]:
        return {ShellMixin.PREFIX: self.shell_enabled}

    def shell_exec(self, code: str) -> tuple[str, str]:
        stdout, stderr = self.do_shell_exec(code)
        log("shell_exec(%s) stdout=%r", code, stdout)
        log("shell_exec(%s) stderr=%r", code, stderr)
        if stdout is not None:
            self.send("shell-reply", 1, stdout)
        if stderr:
            self.send("shell-reply", 2, stderr)
        return stdout, stderr

    def do_shell_exec(self, code) -> tuple[str, str]:
        log("shell_exec(%r)", code)
        if not self.shell_enabled:
            return "shell support is not available with this connection", ""
        try:
            _globals = {
                "connection": self,
                "server": self._server,
                "log": log,
            }
            stdout = io.StringIO()
            stderr = io.StringIO()
            with redirect_stdout(stdout), redirect_stderr(stderr):
                exec(code, _globals, {})  # pylint: disable=exec-used
            return stdout.getvalue(), stderr.getvalue()
        except Exception as e:
            log("shell_exec(..)", exc_info=True)
            log.error("Error running %r:", code)
            log.estr(e)
            return "", str(e)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.gtk.util import get_default_root_window
from tests.xpra.clients.fake_client import FakeClient

Gtk = gi_import("Gtk")


class FakeGTKClient(FakeClient):

    def __init__(self):
        FakeClient.__init__(self)

    def get_mouse_position(self, *_args):
        root = get_default_root_window()
        p = root.get_pointer()
        return p[0], p[1]

    def get_current_modifiers(self, *_args):
        #root = gdk.get_default_root_window()
        #modifiers_mask = root.get_pointer()[-1]
        #return self.mask_to_names(modifiers_mask)
        return []

    def window_close_event(self, *_args):
        Gtk.main_quit()

    def server_ok(self):
        return True


def gtk_main():
    Gtk.main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>


def main(argv=()):
    from xpra.log import Logger, consume_verbose_argv
    from xpra.platform import program_context
    with program_context("opengl", "OpenGL"):
        consume_verbose_argv(argv, "opengl")
        try:
            from xpra.util.system import is_X11
            from xpra.opengl.window import get_gl_client_window_module, test_gl_client_window
            if is_X11():
                from xpra.x11.gtk.display_source import init_gdk_display_source
                init_gdk_display_source()
            opengl_props, gl_client_window_module = get_gl_client_window_module("force")
            log = Logger("opengl")
            log("do_run_glcheck() opengl_props=%s, gl_client_window_module=%s", opengl_props, gl_client_window_module)
            gl_client_window_class = gl_client_window_module.GLClientWindow
            pixel_depth = 0
            log("do_run_glcheck() gl_client_window_class=%s, pixel_depth=%s", gl_client_window_class, pixel_depth)
            # if pixel_depth not in (0, 16, 24, 30) and pixel_depth<32:
            #    pixel_depth = 0
            draw_result = test_gl_client_window(gl_client_window_class, pixel_depth=pixel_depth, show=True)
            success = draw_result.pop("success", False)
            opengl_props.update(draw_result)
            if not success:
                opengl_props["safe"] = False
            return 0
        except Exception:
            log = Logger("opengl")
            log("do_run_glcheck(..)", exc_info=True)
            return 1


if __name__ == "__main__":
    import sys

    r = main(sys.argv)
    sys.exit(r)



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

import os

from xpra.util.io import is_socket
from xpra.util.objects import typedict
from xpra.common import SSH_AGENT_DISPATCH
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.net.ssh.agent import set_ssh_agent, setup_client_ssh_agent_socket, clean_agent_socket
from xpra.log import Logger

log = Logger("network", "ssh")


def accept_client_ssh_agent(uuid: str, ssh_auth_sock: str) -> None:
    sockpath = setup_client_ssh_agent_socket(uuid, ssh_auth_sock)
    if sockpath and os.path.exists(sockpath) and is_socket(sockpath):
        set_ssh_agent(uuid)


class SshAgent(StubServerMixin):
    """
    Mixin for setting up ssh agent forwarding,
    ensures that the symlinks point to the active client,
    and reverts back to the default when the client goes away.
    """
    PREFIX = "ssh-agent"

    def set_session_driver(self, source) -> None:
        if not SSH_AGENT_DISPATCH:
            return
        ssh_auth_sock = getattr(source, "ssh_auth_sock", "")
        set_ssh_agent(ssh_auth_sock)

    def cleanup_protocol(self, protocol) -> None:
        if not SSH_AGENT_DISPATCH:
            return
        source = self.get_server_source(protocol)
        if source and source.uuid:
            clean_agent_socket(source.uuid)
        remaining_sources = tuple(ss for ss in self._server_sources.values() if ss != source)
        for ss in remaining_sources:
            ssh_auth_sock = getattr(ss, "ssh_auth_sock", "")
            if ss.uuid and ssh_auth_sock:
                set_ssh_agent(ss.uuid)
                return
        set_ssh_agent("")

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        if not SSH_AGENT_DISPATCH:
            return
        assert ss
        if ss.uuid and send_ui:
            accept_client_ssh_agent(ss.uuid, c.strget("ssh-auth-sock"))



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.client.gui.window_backing_base import fire_paint_callbacks
from xpra.util.env import envint
from xpra.log import Logger

log = Logger("window", "fake")

FAKE_BACKING_DELAY = envint("XPRA_FAKE_BACKING_DELAY", 5)


class FakeBacking:
    HAS_ALPHA = True

    def __init__(self, wid, *_args):
        self.wid = wid
        self.fake_delay = FAKE_BACKING_DELAY
        self._video_encoder = None
        self._video_encoder_lock = None
        self._video_encoder_speed = []
        self._video_encoder_quality = []

    def close(self):
        self.wid = 0

    def draw_region(self, _x, _y, _width, _height, _coding, _img_data, _rowstride, _options, callbacks):
        log("draw_region(..) faking it after %sms", self.fake_delay)
        from xpra.os_util import gi_import
        GLib = gi_import("GLib")
        GLib.timeout_add(self.fake_delay, fire_paint_callbacks, callbacks, True)

    def cairo_draw(self, context, x, y):
        log("cairo_draw%s", (context, x, y))

    def get_encoding_properties(self):
        return {
            "encodings.rgb_formats": ["RGBA", "RGB"],
        }



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from collections.abc import Sequence

from xpra.util.io import stderr_print
from xpra.util.objects import typedict
from xpra.util.str_fn import obsc
from xpra.auth.sys_auth_base import SysAuthenticatorBase, log
from xpra.auth.common import parse_uid, parse_gid
from xpra.log import is_debug_enabled

assert log  # tests will disable logging from here

LDAP_CACERTFILE = os.environ.get("XPRA_LDAP_CACERTFILE")


class Authenticator(SysAuthenticatorBase):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        self.tls = bool(int(kwargs.pop("tls", "0")))
        self.host = kwargs.pop("host", "localhost")
        self.cacert = kwargs.pop("cacert", LDAP_CACERTFILE)
        self.uid = parse_uid(kwargs.pop("uid", None))
        self.gid = parse_gid(kwargs.pop("gid", None))
        self.tls_version = None
        self.tls_validate = None
        if self.tls:
            import ssl
            tls_version = kwargs.pop("ssl-version", "TLS")
            tls_validate = kwargs.pop("ssl-validate", "REQUIRED")
            self.tls_version = getattr(ssl, "PROTOCOL_%s" % tls_version)
            self.tls_validate = getattr(ssl, "CERT_%s" % tls_validate)
            default_port = 636
        else:
            default_port = 389
        self.port = int(kwargs.pop("port", default_port))
        self.authentication = kwargs.pop("authentication", "NTLM").upper()
        assert self.authentication in ("SIMPLE", "SASL", "NTLM"), \
            "invalid authentication mechanism '%s'" % self.authentication
        super().__init__(**kwargs)
        log("ldap auth: host=%s, port=%i, tls=%s",
            self.host, self.port, self.tls)

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def __repr__(self):
        return "ldap3"

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        self.req_xor(digests)
        return super().get_challenge(("xor", ))

    def check_password(self, password: str) -> bool:
        log("check_password(%s)", obsc(password))
        try:
            from ldap3 import Server, Connection, Tls, ALL, SIMPLE, SASL, NTLM
        except ImportError as e:
            log("check_password(..)", exc_info=True)
            log.warn("Warning: cannot use ldap3 authentication:")
            log.warn(" %s", e)
            return False
        try:
            MECHANISM = {
                "SIMPLE": SIMPLE,
                "SASL": SASL,
                "NTLM": NTLM,
            }
            authentication: SIMPLE | SASL | NTLM = MECHANISM[self.authentication]
            tls = None
            if self.tls:
                tls = Tls(validate=self.tls_validate, version=self.tls_version, ca_certs_file=self.cacert)
                log("TLS=%s", tls)
            server = Server(self.host, port=self.port, tls=tls, use_ssl=self.tls, get_info=ALL)
            log("ldap3 Server(%s)=%s", (self.host, self.port, self.tls), server)
            conn = Connection(server, user=self.username, password=password,
                              authentication=authentication, receive_timeout=10)
            log("ldap3 Connection(%s, %s, %s)=%s", server, self.username, self.authentication, conn)
            if self.tls:
                conn.start_tls()
            r = conn.bind()
            log("ldap3 %s.bind()=%s", conn, r)
            if not r:
                return False
            if is_debug_enabled("auth"):
                log("ldap3 server info:")
                for line in server.info.splitlines():
                    log(f" {line}")
            log("ldap3 who_am_i()=%s", conn.extend.standard.who_am_i())
            return True
        except Exception as e:
            log("ldap3 check_password(..)", exc_info=True)
            log.error("Error: ldap3 authentication failed:")
            log.estr(e)
            return False


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.log import consume_verbose_argv
    from xpra.net.digest import get_salt, get_digests, gendigest
    from xpra.platform import program_context
    with program_context("LDAP3-Password-Auth", "LDAP3-Password-Authentication"):
        consume_verbose_argv(argv, "auth")
        if len(argv) not in (3, 4, 5, 6):
            stderr_print("%s invalid arguments" % argv[0])
            stderr_print("usage: %s username password [host] [port] [tls]" % argv[0])
            return 1
        username = argv[1]
        password = argv[2]
        kwargs = {"username": username}
        if len(argv) >= 4:
            kwargs["host"] = argv[3]
        if len(argv) >= 5:
            kwargs["port"] = argv[4]
        if len(argv) >= 6:
            kwargs["tls"] = argv[5]
        a = Authenticator(**kwargs)
        server_salt, digest = a.get_challenge(("xor", ))
        salt_digest = a.choose_salt_digest(get_digests())
        client_salt = get_salt(len(server_salt))
        combined_salt = gendigest(salt_digest, client_salt, server_salt)
        assert digest == "xor"
        response = gendigest(digest, password, combined_salt)
        caps = typedict({
            "challenge_response": response,
            "challenge_client_salt": client_salt,
        })
        r = a.authenticate(caps)
        print("success: %s" % r)
        return int(not r)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

from typing import Dict, Tuple, Any

from xpra.log import Logger
log = Logger("drm")

cdef extern from "drm.h":
    ctypedef unsigned int __kernel_size_t
    ctypedef struct drm_version:
        int version_major
        int version_minor
        int version_patchlevel
        __kernel_size_t name_len
        char *name
        __kernel_size_t date_len
        char *date
        __kernel_size_t desc_len
        char *desc
    ctypedef drm_version* drmVersionPtr

cdef extern from "xf86drm.h":
    int DRM_NODE_PRIMARY
    int DRM_NODE_CONTROL
    int DRM_NODE_RENDER
    int DRM_NODE_MAX

    ctypedef struct drmDevice:
        char **nodes        #DRM_NODE_MAX sized array
        int available_nodes #DRM_NODE_* bitmask
        int bustype
        #union {
        #drmPciBusInfoPtr pci
        #drmUsbBusInfoPtr usb
        #drmPlatformBusInfoPtr platform
        #drmHost1xBusInfoPtr host1x
        #} businfo;
        #union {
        #drmPciDeviceInfoPtr pci
        #drmUsbDeviceInfoPtr usb
        #drmPlatformDeviceInfoPtr platform
        #drmHost1xDeviceInfoPtr host1x
        #} deviceinfo;
    ctypedef drmDevice* drmDevicePtr

    drmVersionPtr drmGetVersion(int fd)
    void drmFreeVersion(drmVersionPtr ver)

    int drmGetDevices(drmDevicePtr devices[], int max_devices)
    void drmFreeDevices(drmDevicePtr devices[], int count)

#cdef extern from "xf86drmMode.h":
#    int drmIsKMS(int fd)


cdef str s(const char *v):
    pytmp = v[:]
    try:
        return pytmp.decode()
    except:
        return str(v[:])


def get_version() -> Tuple[int, int]:
    return 4, 4


def query() -> Dict[str, Any]:
    info = {}
    cdef int count = drmGetDevices(NULL, 16)
    if count<0:
        log.error(f"Error querying drm devices: {count}")
        return info
    if count==0:
        log.warn("Warning: no drm devices found")
        return info
    cdef drmDevicePtr[16] devices
    count = drmGetDevices(devices, 16)
    assert 0<count<16
    log(f"{count} drm devices found")
    cdef char *path
    cdef int fd
    cdef drmVersionPtr version
    for i in range(count):
        if not devices[i].available_nodes & (1 << DRM_NODE_PRIMARY):
            log(f"{i} is not primary")
            continue
        dev_info = info.setdefault(i, {})
        path = devices[i].nodes[DRM_NODE_PRIMARY]
        log(f"{i} at {s(path)}")
        dev_info["path"] = s(path)
        try:
            with open(path, "rb") as drm_device:
                fd = drm_device.fileno()
                #kms = bool(drmIsKMS(fd))
                #dev_info["kms"] = kms
                version = drmGetVersion(fd)
                if version:
                    dev_info |= {
                        "version"   : (version.version_major, version.version_minor, version.version_patchlevel),
                        "name"      : s(version.name[:version.name_len]),
                        "date"      : s(version.date[:version.date_len]),
                        "desc"      : s(version.desc[:version.desc_len]),
                    }
                    drmFreeVersion(version)
                    #drmModeGetResources
        except OSError as e:
            dev_info["error"] = str(e)
    drmFreeDevices(devices, count)
    return info


def selftest(full=False) -> None:
    info = query()
    log(f"query()={info}")



import win32con        #@UnresolvedImport

print("# win32 constants definitions generated using gen_win32con.py")
print()
for x in dir(win32con):
    v = getattr(win32con, x)
    if x.startswith("_"):
        continue
    if callable(v):
        continue
    if isinstance(v, int):
        print("%s=%s" % (x, v))
    elif isinstance(v, str):
        print("%s='%s'" % (x, v))
    else:
        print("unknown type %s=%s (%s)" % (x, v, type(v)))



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
import struct
from typing import Any

from xpra.common import noop, PaintCallbacks
from xpra.os_util import gi_import, WIN32
from xpra.codecs.loader import load_codec
from xpra.util.objects import typedict
from xpra.platform.paths import get_image, get_image_dir
from xpra.util.io import load_binary_file
from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.gtk.configure.common import sync, run_gui
from xpra.util.config import get_user_config_file, parse_user_config_file, save_user_config_file
from xpra.log import Logger

Gtk = gi_import("Gtk")

log = Logger("opengl", "util")

WW = 640
WH = 480


def BGRA(b: int, g: int, r: int, a: int):
    return struct.pack("@BBBB", b, g, r, a)


def BGR(b: int, g: int, r: int):
    return struct.pack("@BBB", b, g, r)


def draw_data(x: int, y: int, w: int, h: int, pixel_data: bytes):
    coding = "rgb32" if len(pixel_data) == 4 else "rgb24"
    rgb_format = "BGRX" if len(pixel_data) == 4 else "BGR"
    return (
        x, y, w, h,
        coding, pixel_data*w*h, len(pixel_data)*w, {"rgb_format": rgb_format}
    )


CLEAR = (
    "white background",
    (
        draw_data(0, 0, WW, WH, BGRA(0xff, 0xff, 0xff, 0xff)),
    ),
)

TEST_STEPS = [
    CLEAR,
    (
        "a grey square in the top left corner",
        (
            draw_data(0, 0, 128, 128, BGR(0x80, 0x80, 0x80)),
        ),
    ),
    (
        "four squares dividing the screen, red, green, blue and grey",
        (
            draw_data(0, 0, WW//2, WH//2, BGR(0, 0, 0xFF)),
            draw_data(WW//2, 0, WW//2, WH//2, BGR(0, 0xFF, 0)),
            draw_data(WW//2, WH//2, WW//2, WH//2, BGR(0xFF, 0, 0)),
            draw_data(0, WH//2, WW//2, WH//2, BGR(0x40, 0x40, 0x40)),
        ),
    ),
]


def load_image(name: str) -> tuple[int, int, bytes]:
    filename = os.path.join(get_image_dir(), name)
    image_data = load_binary_file(filename)
    w = h = 0
    if image_data:
        try:
            from PIL import Image
            img = Image.open(filename)
            w = int(img.width)
            h = int(img.height)
            img.close()
        except ImportError:
            img = get_image(filename)
            if img:
                w = int(img.get_width())
                h = int(img.get_height())
    return w, h, image_data


def add_test_images():
    encodings = []
    for decoder_name in ("pillow", "jpeg", "webp", "spng", "avif"):
        decoder = load_codec(f"dec_{decoder_name}")
        if decoder:
            encodings += decoder.get_encodings()

    for name, description in {
        "smpte-rp-219.png": "smpte-rp-219 test bars (png)",
        "pinwheel.jpg": "pinwheel pattern (jpeg)",
        "gradient.webp": "gradient pattern (webp)",
    }.items():
        encoding = name.split(".")[-1]
        if encoding not in encodings:
            continue
        width, height, image_data = load_image(name)
        options: dict[str, str] = {}
        if width and height and image_data:
            paint_data = (
                0, 0, width, height,
                encoding, bytes(image_data), 0, options,
            )
            TEST_STEPS.append((description, (paint_data, )))


add_test_images()


def create_twin_test_windows(parent: Gtk.Window) -> tuple[dict[str, Any], list[Gtk.Window]]:
    from xpra.gtk.window import add_close_accel
    from xpra.client.gui.fake_client import FakeClient
    from xpra.client.gui.window_border import WindowBorder
    from xpra.opengl.window import get_gl_client_window_module
    from xpra.client.gtk3.window import ClientWindow
    opengl_props, gl_client_window_module = get_gl_client_window_module("force")
    if not gl_client_window_module:
        return opengl_props, []
    gl_window_class = gl_client_window_module.GLClientWindow
    pixel_depth = 0  # int(opts.pixel_depth)
    wid = 10000
    noclient = FakeClient()
    noclient._id_to_window[wid] = parent
    ww, wh = WW, WH
    metadata = typedict({
        # prevent resizing:
        "maximum-size": (WW, WH),
        "minimum-size": (WW, WH),
        "modal": True,
        "has-alpha": not WIN32,
        "transient-for": wid,
    })
    border = WindowBorder(False)
    max_window_size = None  # (1024, 1024)
    default_cursor_data = None
    windows: list[Gtk.Window] = []
    for i, window_class, title in (
        (0, gl_window_class, "OpenGL Window"),
        (1, ClientWindow, "Non-OpenGL Window"),
    ):
        x, y = 100+ww*i, 100
        window = window_class(noclient, None, 0, 2 ** 32 - 1, x, y, ww, wh, ww, wh,
                              metadata, False, typedict({}),
                              border, max_window_size, default_cursor_data, pixel_depth,
                              )
        window.set_title(title)
        windows.append(window)

        # make sure these windows stay open:
        add_close_accel(window, noop)
    return opengl_props, windows


class ConfigureGUI(BaseGUIWindow):

    def __init__(self, parent: Gtk.Window | None = None):
        self.opengl_props = {}
        self.windows = []
        self.test_steps = TEST_STEPS
        self.step = 0
        super().__init__(
            "Configure Xpra's OpenGL Renderer",
            "opengl.png",
            wm_class=("xpra-configure-opengl-gui", "Xpra Configure OpenGL GUI"),
            header_bar=(False, False),
            parent=parent,
        )

    def dismiss(self, *args):
        self.close_test_windows()
        super().dismiss()

    def populate(self):
        self.populate_form(
            (
                "This tool can cause your system to crash if your GPU drivers are buggy.",
                "Use with caution.",
                "",
                "Enabling the OpenGL renderer can improve the framerate and general performance of the client.",
                "Your xpra client will also be able to skip its OpenGL self-tests and start faster.",
            ),
            ("Proceed", self.start_test),
            ("Exit", self.dismiss),
        )

    def start_test(self, *_args):
        sync()
        self.opengl_props, self.windows = create_twin_test_windows(self)
        if not self.windows:
            log.warn(f"{self.opengl_props=}")
            self.populate_form(
                (
                    "Xpra was unable to intialize the test windows",
                ),
                ("Exit", self.dismiss),
            )
            return
        glp = typedict(self.opengl_props)
        version = ".".join(str(x) for x in glp.inttupleget("opengl", ()))
        renderer = glp.get("renderer", "unknown").split(";")[0]
        backend = glp.get("backend", "unknown")
        vendor = glp.get("vendor", "unknown vendor")
        glinfo = f"OpenGL {version} has been initialized using the {backend!r} backend" + \
            f"and {renderer!r} driver from {vendor}"
        self.populate_form(
            (
                glinfo,
                "",
                "This tool will now present two windows which will be painted using various picture encodings.",
                "You will be asked to confirm that the rendering was correct and identical in both windows.",
                "",
                "Try to arrange them side by side to make it easier to compare.",
            ),
            ("Understood", self.paint_step),
            ("Exit", self.dismiss),
        )

    def paint_step(self, *_args):
        for window in self.windows:
            window.show()
            window.present()
        step_data = self.test_steps[self.step]
        description = step_data[0]
        self.paint_twin_windows(*CLEAR)
        self.paint_twin_windows(*step_data)
        self.populate_form(
            (
                "Please compare the two windows.",
                "They should be indistinguishable from each other.",
                "The colors and content should look 100% identical.",
                "",
                "The windows should be both showing:",
                description,
            ),
            ("Restart", self.restart),
            ("Identical", self.test_passed),
            ("Not identical", self.test_failed),
        )

    def paint_twin_windows(self, description: str, paint_data: tuple):
        log("paint_twin_windows() %r", description)
        callbacks: PaintCallbacks = []
        for x, y, w, h, encoding, img_data, rowstride, options in paint_data:
            for window in self.windows:
                window.draw_region(x, y, w, h, encoding, img_data, rowstride, typedict(options), callbacks)

    def restart(self, *_args):
        self.close_test_windows()
        self.start_test()

    def close_test_windows(self):
        for window in self.windows:
            try:
                window.destroy()
            except Exception:
                log.warn(f"Warning: error closing window {window}", exc_info=True)
        self.windows = []

    def test_passed(self, *_args):
        log("test_passed()")
        self.step += 1
        if self.step < len(self.test_steps):
            self.paint_step()
            return
        # reached the last step!
        self.close_test_windows()
        self.populate_form(
            (
                "OpenGL can be enabled safely using this GPU.",
                "",
                "You can revert this change by running this tool again, or by resetting your user configuration."
            ),
            ("Enable OpenGL", self.enable_opengl),
            ("Exit", self.dismiss),
        )

    def enable_opengl(self, *_args):
        config = parse_user_config_file()
        config["opengl"] = "noprobe"
        save_user_config_file(config)
        self.populate_form(
            (
                "OpenGL is now enabled in your user's xpra configuration file:",
                "'%s'" % get_user_config_file(),
                "If you experience issues later, you may want to reset your configuration.",
            ),
            ("Exit", self.dismiss),
        )

    def test_failed(self, *_args):
        description = self.test_steps[self.step][0]
        self.populate_form(
            (
                "Please report this issue at https://github.com/Xpra-org/xpra/issues/new/choose",
                "",
                f"The test failed on step: {description!r}",
                "Please try to include a screenshot covering both windows",
            ),
            ("Exit", self.dismiss),
        )


def main(_args) -> int:
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os
from typing import Any
from collections.abc import Sequence

from xpra.util.str_fn import csv, print_nested_dict, pver, bytestostr
from xpra.util.env import envbool
from xpra.os_util import POSIX
from xpra.util.io import load_binary_file
from xpra.platform.paths import get_default_conf_dirs, get_system_conf_dirs, get_user_conf_dirs
from xpra.log import Logger, consume_verbose_argv

log = Logger("encoder", "util")

MIN_VERSION = 515

NVIDIA_PROC_FILE = "/proc/driver/nvidia/version"
NVIDIA_HARDWARE = envbool("XPRA_NVIDIA_HARDWARE", False)

nvidia_hardware = 0


def has_nvidia_hardware() -> bool:
    global nvidia_hardware
    if nvidia_hardware == 0:
        nvidia_hardware = _has_nvidia_hardware()
    log(f"has_nvidia_hardware()={nvidia_hardware}")
    return bool(nvidia_hardware)


def _has_nvidia_hardware() -> bool | None:
    if NVIDIA_HARDWARE:
        return True
    # first, check for the kernel module file, this should be very quick:
    try:
        if os.path.exists(NVIDIA_PROC_FILE):
            log(f"has_nvidia_hardware() found kernel module proc file {NVIDIA_PROC_FILE!r}")
            return True
    except OSError:
        log(f"failed to query {NVIDIA_PROC_FILE!r}", exc_info=True)
    # pylint: disable=import-outside-toplevel
    if POSIX:
        # the drm module should also be quick:
        try:
            from xpra.codecs.drm import drm
            info = drm.query()
            for dev_info in info.values():
                dev_name = dev_info.get("name", "").lower()
                if dev_name.find("nouveau") >= 0:
                    continue
                if dev_name.find("nvidia") >= 0:
                    log(f"has_nvidia_hardware() found nvidia drm device: {dev_info}")
                    return True
        except ImportError as e:
            log(f"has_nvidia_hardware() cannot use drm module: {e}")
    try:
        import pynvml
        assert pynvml
        from pynvml import nvmlInit, nvmlShutdown, nvmlDeviceGetCount, NVMLError_DriverNotLoaded
    except ImportError as e:
        log(f"has_nvidia_hardware() cannot use pynvml module: {e}")
    else:
        count = None
        try:
            if nvmlInit():
                count = nvmlDeviceGetCount()
                log(f"has_nvidia_hardware() pynvml found {count} devices")
                return count > 0
        except NVMLError_DriverNotLoaded as e:
            log(f"has_nvidia_hardware() pynvml: {e}")
            return False
        except Exception as e:
            log(f"has_nvidia_hardware() pynvml: {e}")
        finally:
            if count is not None:
                nvmlShutdown()
    # try nvidia-smi for docker contexts
    import subprocess
    try:
        output = subprocess.check_output(["nvidia-smi", "--query-gpu=name", "--format=csv,noheader"])
        output = output.decode("utf-8").strip()
        if output:
            log(f"has_nvidia_hardware() found NVIDIA GPU(s) using nvidia-smi: {output}")
            return True
    except (subprocess.CalledProcessError, OSError):
        log("has_nvidia_hardware() nvidia-smi command not found or failed")
    # hope for the best
    log("has_nvidia_hardware() unable to ascertain, returning None")
    return None


nvml_init_warned = False


def wrap_nvml_init(nvmlInit, warn=True) -> bool:
    try:
        nvmlInit()
        return True
    except Exception as e:
        log("get_nvml_driver_version() pynvml error", exc_info=True)
        global nvml_init_warned
        if not nvml_init_warned and warn:
            log(f"{nvmlInit}()", exc_info=True)
            log.warn("Warning: failed to initialize NVML:")
            log.warn(" %s", e)
            nvml_init_warned = True
        return False


def get_nvml_driver_version() -> Sequence[str]:
    version = ()
    try:
        # pylint: disable=import-outside-toplevel
        from pynvml import nvmlInit, nvmlShutdown, nvmlSystemGetDriverVersion
    except ImportError as e:
        log("cannot use nvml to query the kernel module version:")
        log(" %s", e)
    else:
        try:
            if wrap_nvml_init(nvmlInit):
                try:
                    v = nvmlSystemGetDriverVersion()
                finally:
                    nvmlShutdown()
                log(f"nvmlSystemGetDriverVersion={bytestostr(v)}")
                version = bytestostr(v).split(".")
        except Exception as e:
            log("get_nvml_driver_version() pynvml error", exc_info=True)
            log.warn("Warning: failed to query the NVidia kernel module version using NVML:")
            log.warn(" %s", e)
    log(f"get_nvml_driver_version()={version}")
    return version


def get_proc_driver_version() -> Sequence[str]:
    if not POSIX:
        return ()
    v = load_binary_file(NVIDIA_PROC_FILE)
    if not v:
        log.warn("Warning: NVidia kernel module not installed?")
        log.warn(f" cannot load {NVIDIA_PROC_FILE!r}")
        return ()
    version = ()
    # ie: "NVRM version: NVIDIA UNIX x86_64 Kernel Module  565.77  Wed Nov 27 23:33:08 UTC 2024"
    # or: "NVIDIA UNIX Open Kernel Module for x86_64  575.57.08  Release Build  (dvs-builder@U22-I3-H04-01-5)  ..."
    KSTR = "Kernel Module"
    for line in bytestostr(v).splitlines():
        if line.find(KSTR) < 0:
            continue
        # try to split on double space:
        parts = line.split("  ")
        if len(parts) >= 3:
            version = parts[1].split(".")
            break
    if version:
        log_fn = log.debug
    else:
        log_fn = log.warn
        log_fn("Warning: unable to parse NVidia kernel module version")
    log_fn(f" {NVIDIA_PROC_FILE!r} contents:")
    for line in v.splitlines():
        log_fn(f"   {bytestostr(line)!r}")
    log(f"get_proc_driver_version()={version}")
    return version


def identify_nvidia_module_version() -> Sequence[int]:
    v = get_nvml_driver_version() or get_proc_driver_version()
    # only keep numeric values:
    numver = []
    try:
        for x in v:
            try:
                numver.append(int(x))
            except ValueError:
                if not numver:
                    raise
        if numver:
            log.info("NVidia driver version %s", pver(numver))
            return tuple(numver)
    except Exception as e:
        log.warn(f"Warning: failed to parse Nvidia driver version {v!r}: {e}")
    return ()


nvidia_module_version = None


def get_nvidia_module_version(probe=True) -> Sequence[int] | None:
    global nvidia_module_version
    if nvidia_module_version is None and probe:
        nvidia_module_version = identify_nvidia_module_version()
    return nvidia_module_version


def identify_cards() -> dict:
    devices: dict[int, dict[str, Any]] = {}
    try:
        # pylint: disable=import-outside-toplevel
        import pynvml
        from pynvml import nvmlInit, nvmlShutdown, nvmlDeviceGetCount, nvmlDeviceGetHandleByIndex
        deviceCount = None
        try:
            if not wrap_nvml_init(nvmlInit):
                return devices
            deviceCount = nvmlDeviceGetCount()
            log(f"identify_cards() will probe {deviceCount} cards")
            for i in range(deviceCount):
                handle = nvmlDeviceGetHandleByIndex(i)
                log(f"identify_cards() handle({i})={handle}")
                props: dict[str, Any] = {}

                def meminfo(memory):
                    return {
                        "total": int(memory.total),
                        "free": int(memory.free),
                        "used": int(memory.used),
                    }

                def pciinfo(pci) -> dict[str, int | str]:
                    i = {}
                    for nvname, pubname in {
                        "domain": "domain",
                        "bus": "bus",
                        "device": "device",
                        "pciDeviceId": "pci-device-id",
                        "pciSubSystemId": "pci-subsystem-id",
                    }.items():
                        try:
                            i[pubname] = int(getattr(pci, nvname))
                        except (ValueError, AttributeError):
                            pass
                    try:
                        i["bus-id"] = bytestostr(pci.busId)
                    except AttributeError:
                        pass
                    return i

                for prefix, prop, fn_name, args, conv in (
                        ("", "name", "nvmlDeviceGetName", (), bytestostr),
                        ("", "serial", "nvmlDeviceGetSerial", (), bytestostr),
                        ("", "uuid", "nvmlDeviceGetUUID", (), bytestostr),
                        ("", "pci", "nvmlDeviceGetPciInfo", (), pciinfo),
                        ("", "memory", "nvmlDeviceGetMemoryInfo", (), meminfo),
                        ("pcie-link", "generation-max", "nvmlDeviceGetMaxPcieLinkGeneration", (), int),
                        ("pcie-link", "width-max", "nvmlDeviceGetMaxPcieLinkWidth", (), int),
                        ("pcie-link", "generation", "nvmlDeviceGetCurrPcieLinkGeneration", (), int),
                        ("pcie-link", "width", "nvmlDeviceGetCurrPcieLinkWidth", (), int),
                        ("clock-info", "graphics", "nvmlDeviceGetClockInfo", (0,), int),
                        ("clock-info", "sm", "nvmlDeviceGetClockInfo", (1,), int),
                        ("clock-info", "mem", "nvmlDeviceGetClockInfo", (2,), int),
                        ("clock-info", "graphics-max", "nvmlDeviceGetMaxClockInfo", (0,), int),
                        ("clock-info", "sm-max", "nvmlDeviceGetMaxClockInfo", (1,), int),
                        ("clock-info", "mem-max", "nvmlDeviceGetMaxClockInfo", (2,), int),
                        ("", "fan-speed", "nvmlDeviceGetFanSpeed", (), int),
                        ("", "temperature", "nvmlDeviceGetTemperature", (0,), int),
                        ("", "power-state", "nvmlDeviceGetPowerState", (), int),
                        ("", "vbios-version", "nvmlDeviceGetVbiosVersion", (), bytestostr),
                ):
                    try:
                        fn = getattr(pynvml, fn_name)
                    except AttributeError:
                        log(f"{fn_name} not found in {pynvml}")
                        continue
                    try:
                        v = fn(handle, *args)
                        if conv:
                            v = conv(v)
                        if prefix:
                            d = props.setdefault(prefix, {})
                        else:
                            d = props
                        d[prop] = v
                    except Exception as e:
                        log("identify_cards() cannot query %s using %s on device %i with handle %s: %s",
                            prop, fn, i, handle, e)
                        continue
                log(f"identify_cards() [{i}]={props}")
                devices[i] = props
            # unitCount = nvmlUnitGetCount()
            # log.info("unitCount=%s", unitCount)
        except Exception as e:
            log("identify_cards() pynvml error", exc_info=True)
            log.warn("Warning: failed to query the NVidia cards using NVML:")
            log.warn(" %s", e)
        finally:
            if deviceCount is not None:
                nvmlShutdown()
    except ImportError as e:
        log("cannot use nvml to query the kernel module version:")
        log(" %s", e)
    return devices


_cards = None


def get_cards(probe=True) -> dict:
    global _cards
    if _cards is None and probe:
        _cards = identify_cards()
    return _cards


def is_blocklisted() -> bool | None:
    v = get_nvidia_module_version(True)
    if v:
        try:
            if v[0] > MIN_VERSION:
                return False
        except Exception as e:
            log("is_blocklisted()", exc_info=True)
            log.warn(f"Warning: error checking driver version {v!r}:")
            log.warn(" %s", e)
    return None  # we don't know: unreleased / untested


_version_warning = False


def validate_driver_yuv444lossless() -> bool:
    # this should log the kernel module version
    v = get_nvidia_module_version()
    if not v:
        log.warn("Warning: unknown NVidia driver version")
        bl = None
    else:
        bl = is_blocklisted()
    if bl is True:
        raise ValueError(f"NVidia driver version {pver(v)} is blocklisted, it does not work with NVENC")
    if bl is None:
        global _version_warning
        if _version_warning:
            log_fn = log.debug
        else:
            log_fn = log.warn
            _version_warning = True
        if v[0] < MIN_VERSION:
            log_fn(f"Warning: NVidia driver version {pver(v)} is untested with NVENC")
            log_fn(f" (this encoder has been tested with versions {MIN_VERSION}.x and later only)")
        if not envbool("XPRA_NVENC_YUV444P", False):
            log_fn(" disabling YUV444P and lossless mode")
            log_fn(" use XPRA_NVENC_YUV444P=1 to force enable")
            return False
        log_fn(" force enabling YUV444P and lossless mode")
    return True


def parse_nvfbc_hex_key(s) -> bytes:
    # ie: 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10
    # ie: 0102030405060708090A0B0C0D0E0F10
    # start by removing spaces and 0x:
    hexstr = s.replace("0x", "").replace(",", "").replace(" ", "")
    import binascii  # pylint: disable=import-outside-toplevel
    return binascii.unhexlify(hexstr)


license_keys: dict[str, tuple] = {}


def get_license_keys(version=0, basefilename="nvenc") -> Sequence[str]:
    filename = f"{basefilename}%s.keys" % (version or "")
    keys = license_keys.get(filename)
    if keys is not None:
        return keys
    env_name = f"XPRA_{basefilename.upper()}_CLIENT_KEY"
    env_keys = os.environ.get(env_name, "")
    if env_keys:
        keys = [x.strip() for x in env_keys.split(",")]
        log(f"using {basefilename} keys from environment variable {env_name} : {csv(keys)}")
    else:
        # try to load the license file
        keys = []
        with log.trap_error(f"Error loading {basefilename!r} license keys"):
            # see read_xpra_defaults for an explanation of paths
            dirs = get_default_conf_dirs() + get_system_conf_dirs() + get_user_conf_dirs()
            for d in dirs:
                if not d:
                    continue
                keys_file = os.path.join(d, filename)
                keys_file = os.path.expanduser(keys_file)
                if not os.path.exists(keys_file):
                    log(f"get_license_keys({basefilename}, {version}) {keys_file!r} does not exist")
                    continue
                log(f"loading {basefilename} version {version} keys from {keys_file!r}")
                with open(keys_file, "rb") as f:
                    fkeys = []
                    for line in f:
                        sline = line.strip().rstrip(b'\r\n').strip().decode("latin1")
                        if not sline:
                            log("skipping empty line")
                            continue
                        if sline[0] in ('!', '#'):
                            log("skipping comments")
                            continue
                        fkeys.append(sline)
                        log(f"added key: {sline}")
                    log(f"added {len(fkeys)} keys from {keys_file}")
                    keys += fkeys
    license_keys[filename] = keys
    log(f"get_nvenc_license_keys({version})={keys}")
    return keys


def main():
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("Nvidia-Info", "Nvidia Info"):
        consume_verbose_argv(sys.argv, "encoding")
        # this will log the version number:
        get_nvidia_module_version()
        if is_blocklisted():
            log.warn("Warning: this driver version is blocklisted")
        keys = get_license_keys()
        log.info(f"{len(keys)} NVENC license keys")
        for k in keys:
            log.info(f"  {k}")
        try:
            import pynvml
            assert pynvml
        except ImportError:
            log.warn("Warning: the pynvml library is missing")
            log.warn(" cannot identify the GPUs installed")
        else:
            cards = get_cards()
            if cards:
                log.info("")
                log.info(f"{len(cards)} cards:")
                print_nested_dict(cards, print_fn=log.info)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=line-too-long

import unittest

from xpra.net.digest import (
    get_digests, get_digest_module,
    choose_digest, gendigest, verify_digest,
    get_salt,
)


class TestDigest(unittest.TestCase):

    def test_invalid_digest(self):
        for invalid_digest in (None, "foo", "hmac", "hmac+INVALID_HASH_ALGO"):
            try:
                assert get_digest_module(invalid_digest) is None
            except TypeError:
                pass
            try:
                assert not gendigest(invalid_digest, "bar", b"0"*16)
            except TypeError:
                pass

    # noinspection PyTypeChecker
    def test_all_digests(self):
        for digest in get_digests():
            if digest.startswith("hmac"):
                m = get_digest_module(digest)
                assert m is not None, "digest module not found for '%s'" % digest
            salt = get_salt(32)
            password = "secret"
            d = gendigest(digest, password, salt)
            assert d is not None

            def nvd(password=password, salt=salt, response=d):
                try:
                    r = verify_digest(digest, password, salt, response)
                    assert not r
                except TypeError:
                    pass
            nvd(password=None)
            nvd(salt=None)
            nvd(response=None)
            # truncated:
            nvd(password=password[1:])
            nvd(salt=salt[1:])
            nvd(response=d[1:])
            assert verify_digest(digest, password, salt, d)

    def test_choose_digest(self):
        for h in ("hmac+sha512", "hmac+sha384", "hmac+sha256", "hmac+sha224",
                  "xor", "des"):
            assert choose_digest((h,))==h
            assert choose_digest((h, "hmac+sha512"))=="hmac+sha512"


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.auth.sys_auth_base import SysAuthenticator


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        self.var_name = kwargs.pop("name", "XPRA_PASSWORD")
        super().__init__(**kwargs)
        self.authenticate_check = self.authenticate_hmac

    def __repr__(self):
        return "env"

    def get_password(self) -> str:
        return os.environ.get(self.var_name, "")



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.x11.common import Unmanageable
from xpra.x11.models.base import BaseWindowModel
from xpra.x11.bindings.window import X11WindowBindings

X11Window = X11WindowBindings()

GObject = gi_import("GObject")


class OverrideRedirectWindowModel(BaseWindowModel):
    __gsignals__ = dict(BaseWindowModel.__common_signals__)
    __gproperties__ = dict(BaseWindowModel.__common_properties__)
    __gproperties__ |= {
        "override-redirect": (
            GObject.TYPE_BOOLEAN,
            "Is the window of type override-redirect", "",
            True,
            GObject.ParamFlags.READABLE,
        ),
    }
    _property_names = BaseWindowModel._property_names + ["override-redirect"]
    _MODELTYPE = "OR-Window"

    def __init__(self, xid: int):
        super().__init__(xid)
        self._updateprop("override-redirect", True)

    def setup(self) -> None:
        super().setup()
        # So now if the window becomes unmapped in the future then we will
        # notice... but it might be unmapped already, and any event
        # already generated, and our request for that event is too late!
        # So double check now, *after* putting in our request:
        if not X11Window.is_mapped(self.xid):
            raise Unmanageable("window already unmapped")
        ch = self._composite.get_contents_handle()
        if ch is None:
            raise Unmanageable("failed to get damage handle")


GObject.type_register(OverrideRedirectWindowModel)



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# posix: assume cups support (no overrides needed here)
from xpra.platform.pycups_printing import (
    get_printers,
    print_files,
    printing_finished,
    init_printing,
    cleanup_printing,
    get_info,
)

for x in (get_printers, print_files, printing_finished, init_printing, cleanup_printing, get_info):
    assert callable(x), f"{x} is not callable"



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.client.gui.keyboard_helper import KeyboardHelper, log
from xpra.gtk.keymap import get_gtk_keymap, get_default_keymap
from xpra.os_util import gi_import
from xpra.util.system import is_X11

Gdk = gi_import("Gdk")
GLib = gi_import("GLib")


class GTKKeyboardHelper(KeyboardHelper):

    def __init__(self, *args, **kwargs):
        KeyboardHelper.__init__(self, *args, **kwargs)
        # used for delaying the sending of keymap changes
        # (as we may be getting dozens of such events at a time)
        self._keymap_changing = False
        self._keymap_change_handler_id = None
        self._keymap = get_default_keymap()
        self.update()
        if self._keymap:
            self._keymap_change_handler_id = self._keymap.connect("keys-changed", self.keymap_changed)

    def next_layout(self, update_platform_layout) -> None:
        log(f"next_layout(update_platform_layout={update_platform_layout})")
        if self.layout_option not in self.layouts_option:
            log("no layout change; use --keyboard-layout/--keyboard-layouts to specify the layouts order")
            return
        try:
            layout_index = self.layouts_option.index(self.layout_option)
        except ValueError as e:
            log.warn("failed to find layout %s among layouts: %s", self.layout_option, e)
            return
        layout_index = (layout_index + 1) % len(self.layouts_option)
        self.layout_option = self.layouts_option[layout_index]
        log.info("calling keymap_changed to apply %s layout", self.layout_option)
        self.keymap_changed()
        if update_platform_layout:
            log("updating the platform layout to %s", self.layout_option)
            self.set_platform_layout(self.layout_option)

    def keymap_changed(self, *args) -> None:
        log("keymap_changed%s", args)
        if self._keymap_change_handler_id:
            self._keymap.disconnect(self._keymap_change_handler_id)
            self._keymap_change_handler_id = None
        self._keymap = get_default_keymap()
        if self._keymap_changing:
            # timer is already due
            return
        self._keymap_changing = True

        def do_keys_changed() -> None:
            # re-register the change handler:
            self._keymap_change_handler_id = self._keymap.connect("keys-changed", self.keymap_changed)
            self._keymap_changing = False
            if self.locked:
                # automatic changes not allowed!
                log.info("ignoring keymap change: layout is locked to '%s'", self.layout_str())
                return
            if self.update() and self.layout:
                log.info("keymap has been changed to '%s'", self.layout_str())
                log.info(" sending updated mappings to the server")
                if self.layout:
                    self.send_layout()
                self.send_keymap()

        GLib.timeout_add(500, do_keys_changed)

    def update(self) -> bool:
        old_hash = self.hash
        super().update()
        if is_X11():
            with log.trap_error("Error querying modifier map"):
                self.keyboard.update_modifier_map(self.mod_meanings)
        log("update() modifier_map=%s, old hash=%s, new hash=%s", self.keyboard.modifier_map, old_hash, self.hash)
        return old_hash != self.hash

    def get_full_keymap(self) -> Sequence[tuple[int, str, int, int, int]]:
        return get_gtk_keymap()

    def cleanup(self) -> None:
        super().cleanup()
        if self._keymap_change_handler_id:
            try:
                self._keymap.disconnect(self._keymap_change_handler_id)
                self._keymap_change_handler_id = None
            except Exception as e:
                log.warn("failed to disconnect keymap change handler: %s", e)


def main() -> None:
    # use gtk as display source:
    # pylint: disable=import-outside-toplevel
    from xpra.common import noop
    from xpra.gtk.util import init_display_source
    from xpra.util.str_fn import print_nested_dict
    from xpra.platform import program_context
    with program_context("GTK-Keyboard", "GTK Keyboard"):
        init_display_source()
        x = GTKKeyboardHelper(noop)
        x.query_xkbmap()
        print_nested_dict(x.get_keymap_properties())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import gi

gi.require_version("Gtk", "3.0")  # @UndefinedVariable
from gi.repository import Gtk  # pylint: disable=wrong-import-position @UnresolvedImport
from xpra.x11.gtk.display_source import init_gdk_display_source

init_gdk_display_source()
from xpra.x11.gtk.prop import prop_set


def main():
    window = Gtk.Window()
    window.set_size_request(220, 120)
    window.connect("delete_event", Gtk.main_quit)
    vbox = Gtk.VBox(homogeneous=False, spacing=0)

    b = Gtk.Button(label="Bypass")

    def bypass(*args):
        prop_set(window.get_window(), "_NET_WM_BYPASS_COMPOSITOR", "u32", 1)

    b.connect('clicked', bypass)
    vbox.add(b)

    b = Gtk.Button(label="Not Bypass")

    def notbypass(*args):
        prop_set(window.get_window(), "_NET_WM_BYPASS_COMPOSITOR", "u32", 2)

    b.connect('clicked', notbypass)
    vbox.add(b)

    window.add(vbox)
    window.show_all()
    Gtk.main()
    return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import logging
import binascii

from xpra.challenge.handler import AuthenticationHandler
from xpra.util.env import osexpand
from xpra.util.io import load_binary_file
from xpra.util.str_fn import strtobytes
from xpra.log import Logger, is_debug_enabled

log = Logger("auth")


class Handler(AuthenticationHandler):

    def __init__(self, **kwargs):
        self.protocol = kwargs["protocol"]
        self.app_id = kwargs.get("APP_ID", "") or os.environ.get("XPRA_U2F_APP_ID", "") or "Xpra"

    def __repr__(self):
        return "u2f"

    def get_digest(self) -> str:
        return "u2f"

    def handle(self, challenge: bytes, digest: str, prompt: str) -> (
            tuple[bytes, bytes] | None):  # pylint: disable=unused-argument
        if not digest.startswith("u2f:"):
            log("%s is not a u2f challenge", digest)
            return None
        try:
            # pylint: disable=import-outside-toplevel
            from pyu2f import model
            from pyu2f.u2f import GetLocalU2FInterface
        except ImportError as e:
            log.warn("Warning: cannot use u2f authentication handler")
            log.warn(" %s", e)
            return None
        if not is_debug_enabled("auth"):
            logging.getLogger("pyu2f.hardware").setLevel(logging.INFO)
            logging.getLogger("pyu2f.hidtransport").setLevel(logging.INFO)
        dev = GetLocalU2FInterface()
        key_handle = self.get_key_handle()
        if not key_handle:
            return None
        key = model.RegisteredKey(key_handle)
        # use server salt as challenge directly
        log.info("activate your U2F device for authentication")
        log.info(f"prompt: {prompt!r}")
        response = dev.Authenticate(self.app_id, challenge, [key])
        sig = response.signature_data
        client_data = response.client_data
        log("process_challenge_u2f client data=%s, signature=%s", client_data, binascii.hexlify(sig))
        return bytes(sig), strtobytes(client_data.origin)

    def get_key_handle(self) -> bytes:
        key_handle_str = os.environ.get("XPRA_U2F_KEY_HANDLE")
        log("process_challenge_u2f XPRA_U2F_KEY_HANDLE=%s", key_handle_str)
        if not key_handle_str:
            # try to load the key handle from the user conf dir(s):
            from xpra.platform.paths import get_user_conf_dirs  # pylint: disable=import-outside-toplevel
            info = self.protocol.get_info(False)
            key_handle_filenames = []
            for hostinfo in ("-%s" % info.get("host", ""), ""):
                for d in get_user_conf_dirs():
                    key_handle_filenames.append(os.path.join(d, f"u2f-keyhandle{hostinfo}.hex"))
            for filename in key_handle_filenames:
                p = osexpand(filename)
                key_handle_str = load_binary_file(p).rstrip(b" \n\r").decode("latin1")
                log("key_handle_str(%s)=%s", p, key_handle_str)
                if key_handle_str:
                    break
            if not key_handle_str:
                log.warn("Warning: no U2F key handle found")
                return b""
        key_handle = binascii.unhexlify(key_handle_str)
        log("process_challenge_u2f key_handle=%s", key_handle)
        return key_handle



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import subprocess
from collections.abc import Callable, Iterable

from xpra.platform.win32 import constants as win32con
from xpra.util.objects import reverse_dict
from xpra.util.str_fn import csv, bytestostr
from xpra.util.env import envint, envbool
from xpra.platform.paths import get_app_dir
from xpra.log import Logger

log = Logger("printing")

# allows us to skip some printers we don't want to export
WINSPOOL_LISTENER = envbool("XPRA_WINSPOOL_LISTENER", True)
SKIPPED_PRINTERS = os.environ.get("XPRA_SKIPPED_PRINTERS", "Microsoft XPS Document Writer,Fax").split(",")
PRINTER_LEVEL = envint("XPRA_WIN32_PRINTER_LEVEL", 1)
DEFAULT_PRINTER_FLAGS = "LOCAL,SHARED+NETWORK+CONNECTIONS"
PRINTER_FLAGS = [x.strip() for x in os.environ.get("XPRA_WIN32_PRINTER_FLAGS", DEFAULT_PRINTER_FLAGS).split(",")]

DEFAULT_MIMETYPES = ["application/pdf", ]

PRINTER_ENUM_VALUES = {
    "DEFAULT": 1,
    "LOCAL": 2,
    "CONNECTIONS": 4,
    "NAME": 8,
    "REMOTE": 16,
    "SHARED": 32,
    "NETWORK": 64,
    "EXPAND": 16384,
    "CONTAINER": 32768,
    "ICON1": 65536 * 1,
    "ICON2": 65536 * 2,
    "ICON3": 65536 * 3,
    "ICON4": 65536 * 4,
    "ICON5": 65536 * 5,
    "ICON6": 65536 * 6,
    "ICON7": 65536 * 7,
    "ICON8": 65536 * 8,
}
PRINTER_ENUM_NAMES = reverse_dict(PRINTER_ENUM_VALUES)
log("PRINTER_ENUM_VALUES: %s", PRINTER_ENUM_VALUES)

log("PRINTER_FLAGS=%s", csv(PRINTER_FLAGS))
VALID_PRINTER_FLAGS = ("LOCAL", "SHARED", "CONNECTIONS", "NETWORK", "REMOTE")


def get_printer_enums():
    printer_enums = []
    for v in PRINTER_FLAGS:  # ie: "SHARED+NETWORK+CONNECTIONS"
        flags = v.replace('|', '+').split("+")  # ie: ["SHARED", "NETWORK", "CONNECTIONS"]
        values = []
        for flag in flags:  # ie: "SHARED"
            if flag not in VALID_PRINTER_FLAGS:
                log.warn("Warning: the following printer flag is invalid and will be ignored: %s", flag)
            else:
                values.append(flag)  # ie: "SHARED"
        printer_enums.append(values)
    return tuple(printer_enums)


PRINTER_ENUMS = get_printer_enums()
log("PRINTER_ENUMS=%s", PRINTER_ENUMS)

# emulate pycups job id
JOB_ID = 0
PROCESSES = {}

printers_modified_callback: Callable | None = None


def init_printing(callback=None):
    global printers_modified_callback
    log("init_printing(%s) printers_modified_callback=%s", callback, printers_modified_callback)
    printers_modified_callback = callback
    if WINSPOOL_LISTENER:
        try:
            init_winspool_listener()
        except Exception:
            log.error("Error: failed to register for print spooler changes", exc_info=True)


def init_winspool_listener():
    from xpra.platform.win32.events import get_win32_event_listener
    get_win32_event_listener().add_event_callback(win32con.WM_DEVMODECHANGE, on_devmodechange)


def on_devmodechange(wParam, lParam):
    global printers_modified_callback
    log("on_devmodechange(%s, %s) printers_modified_callback=%s", wParam, lParam, printers_modified_callback)
    # from ctypes import c_wchar_p
    # name = c_wchar_p(lParam)
    # log("device changed: %s", name)
    if lParam > 0 and printers_modified_callback:
        printers_modified_callback()


def EnumPrinters(flags, name=None, level=PRINTER_LEVEL):
    from ctypes import (
        WinDLL,  # @UnresolvedImport
        byref, addressof, cast, Structure, POINTER,
    )
    from ctypes.wintypes import BYTE, BOOL, DWORD, LPCWSTR, LPBYTE, LPDWORD

    winspool = WinDLL('winspool.drv', use_last_error=True)
    EnumPrintersW = winspool.EnumPrintersW
    EnumPrintersW.restype = BOOL
    EnumPrintersW.argtypes = [DWORD, LPCWSTR, DWORD, LPBYTE, DWORD, LPDWORD, LPDWORD]

    class PRINTER_INFO(Structure):
        _fields_ = [
            ("Flags", DWORD),
            ("pDescription", LPCWSTR),
            ("pName", LPCWSTR),
            ("pComment", LPCWSTR),
        ]

    # Invoke once with a NULL pointer to get buffer size.
    info = BYTE(0)
    pcbNeeded = DWORD(0)
    pcReturned = DWORD(0)  # the number of PRINTER_INFO_1 structures retrieved
    r = EnumPrintersW(DWORD(flags), name, DWORD(level), byref(info), DWORD(0), byref(pcbNeeded), byref(pcReturned))
    log("EnumPrintersW(..)=%i pcbNeeded=%i", r, pcbNeeded.value)
    if pcbNeeded.value <= 0:
        log("EnumPrinters probe failed for flags=%i, level=%i, pcbNeeded=%i", flags, level, pcbNeeded.value)
        return []

    bufsize = int(pcbNeeded.value)
    # noinspection PyCallingNonCallable,PyTypeChecker
    buf = (BYTE * bufsize)()
    pPrinterEnum = cast(addressof(buf), LPBYTE)
    log("EnumPrinters into buf %s", pPrinterEnum)
    r = EnumPrintersW(DWORD(flags), name, DWORD(level), pPrinterEnum, DWORD(bufsize), byref(pcbNeeded),
                      byref(pcReturned))
    log("EnumPrintersW(..)=%i pcReturned=%i", r, pcReturned.value)
    if r == 0:
        log.error("Error: EnumPrinters failed")
        return []
    info = cast(pPrinterEnum, POINTER(PRINTER_INFO))
    printers = []
    for i in range(pcReturned.value):
        v = int(info[i].Flags), str(info[i].pDescription), str(info[i].pName), str(info[i].pComment)
        log("EnumPrintersW(..) [%i]=%s", i, v)
        printers.append(v)
    del buf
    return printers


def get_info():
    from xpra.platform.printing import default_get_info
    i = default_get_info()
    # win32 extras:
    i.update({
        "skipped-printers": SKIPPED_PRINTERS,
        "printer-level": PRINTER_LEVEL,
        "default-printer-flags": DEFAULT_PRINTER_FLAGS,
        "printer-flags": PRINTER_FLAGS,
    })
    return i


def get_printers():
    global PRINTER_ENUMS, PRINTER_ENUM_VALUES, SKIPPED_PRINTERS, PRINTER_LEVEL
    printers = {}
    for penum in PRINTER_ENUMS:
        try:
            eprinters = []
            enum_values = [PRINTER_ENUM_VALUES.get(x, 0) for x in penum]
            enum_val = sum(enum_values)
            log("enum(%s)=%s=%s", penum, "+".join(str(x) for x in enum_values), enum_val)
            assert enum_val is not None, "invalid printer enum %s" % penum
            log("querying %s printers with level=%s", penum, PRINTER_LEVEL)
            for p in EnumPrinters(enum_val, None, PRINTER_LEVEL):
                flags, desc, name, comment = p
                if name in SKIPPED_PRINTERS:
                    log("skipped printer: %#x, %s, %s, %s", flags, desc, name, comment)
                    continue
                if name in printers:
                    log("skipped duplicate printer: %#x, %s, %s, %s", flags, desc, name, comment)
                    continue
                log("found printer: %#x, %s, %s, %s", flags, desc, name, comment)
                # strip duplicated and empty strings from the description:
                desc_els = []
                for x in desc.split(","):
                    if x and not desc_els.count(x):
                        desc_els.append(x)
                info = {"printer-info": bytestostr(",".join(desc_els)),
                        "type": penum}
                if comment:
                    info["printer-make-and-model"] = comment
                printers[name] = info
                eprinters.append(name)
            log("%s printers: %s", penum, eprinters)
        except Exception as e:
            log.warn("Warning: failed to query %s printers:", penum)
            log.warn(" %s", e)
            log("query error", exc_info=True)
            del e
    log("win32.get_printers()=%s", printers)
    return printers


def print_files(printer: str, filenames: Iterable[str], title: str, options: dict):
    log("win32.print_files%s", (printer, filenames, title, options))
    global JOB_ID, PROCESSES
    processes = []
    for filename in filenames:
        cwd = get_app_dir()
        command = ["PDFIUM_Print.exe", filename, printer, title]
        log("print command: %s", command)
        startupinfo = subprocess.STARTUPINFO()  # @UndefinedVariable
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # @UndefinedVariable
        startupinfo.wShowWindow = 0  # aka win32.con.SW_HIDE
        process = subprocess.Popen(command, stdout=sys.stdout, stderr=sys.stderr, cwd=cwd, startupinfo=startupinfo)
        process.print_filename = filename
        # we just let it run, no need for reaping the process on win32
        processes.append(process)
    JOB_ID += 1
    PROCESSES[JOB_ID] = processes
    log("win32.print_files(..)=%s (%s)", JOB_ID, processes)
    return JOB_ID


def printing_finished(jobid):
    global PROCESSES
    processes = PROCESSES.get(jobid)
    if not processes:
        log.warn("win32.printing_finished(%s) job not found!", jobid)
        return True
    log("win32.printing_finished(%s) processes: %s", jobid, [x.print_filename for x in processes])
    pending = [proc.print_filename for proc in processes if proc.poll() is None]
    log("win32.printing_finished(%s) still pending: %s", jobid, pending)
    # return finished when all the processes have terminated
    return len(pending) == 0



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import signal
from time import sleep, monotonic

from xpra import __version__
from xpra.util.env import envint, envbool, ignorewarnings
from xpra.os_util import OSX, WIN32, gi_import
from xpra.util.system import SIGNAMES
from xpra.exit_codes import ExitCode, ExitValue
from xpra.common import SPLASH_EXIT_DELAY
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.signals import install_signal_handlers
from xpra.gtk.css_overrides import inject_css_overrides
from xpra.platform.gui import force_focus, set_window_progress
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")

log = Logger("client", "util")

inject_css_overrides()

TIMEOUT = envint("XPRA_SPLASH_TIMEOUT", 60)
LINES = envint("XPRA_SPLASH_LINES", 4)
READ_SLEEP = envint("XPRA_SPLASH_READ_SLEEP", 0)
FOCUS_EXIT = envbool("XPRA_SPLASH_FOCUS_EXIT", True)

# alternative: "▁▂▃▄▅▆▇█▇▆▅▄▃▁"
PULSE_CHARS = "◐◓◑◒"
if OSX:
    DONE_CHAR = "-"
else:
    DONE_CHAR = "⬤"

W = 400


class SplashScreen(Gtk.Window):

    def __init__(self):
        self.exit_code = None
        super().__init__(type=Gtk.WindowType.TOPLEVEL)
        self.connect("delete_event", self.exit)
        self.add_events(Gdk.EventType.BUTTON_PRESS)
        self.connect("button-press-event", self.exit)
        self.connect("key-press-event", self.exit)
        title = "Xpra %s" % __version__
        self.set_title(title)
        self.set_size_request(W, 40 + 40 * LINES)
        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_decorated(False)
        import warnings
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        self.set_opacity(0.9)
        self.set_accept_focus(False)
        self.set_focus_on_map(False)
        self.set_skip_pager_hint(True)
        self.set_skip_taskbar_hint(True)
        if not OSX:
            self.set_type_hint(Gdk.WindowTypeHint.SPLASHSCREEN)
        vbox = Gtk.VBox()
        hbox = Gtk.HBox(homogeneous=False)
        icon = get_icon_pixbuf("xpra.png")
        if icon:
            self.set_icon(icon)
            hbox.pack_start(Gtk.Image.new_from_pixbuf(icon), False, False, 20)
        self.title_label = label(title, font="sans 18")
        hbox.pack_start(self.title_label, True, True, 20)
        vbox.add(hbox)
        self.labels = []
        for i in range(LINES):
            lbl = label(" ")
            lbl.set_opacity((i + 1) / LINES)
            self.labels.append(lbl)
            al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0, yscale=0)
            al.add(lbl)
            vbox.pack_start(al, True, True, 4)
        self.progress_bar = Gtk.ProgressBar()
        self.progress_bar.set_size_request(320, 30)
        self.progress_bar.set_show_text(False)
        self.progress_bar.set_fraction(0)
        self.progress_timer = 0
        self.fade_out_timer = 0
        self.exit_timer = 0
        vbox.add(self.progress_bar)
        self.timeout_timer = 0
        self.pulse_timer = 0
        self.pulse_counter = 0
        self.current_label_text = None
        self.add(vbox)
        install_signal_handlers("", self.handle_signal)
        sigpipe = getattr(signal, "SIGPIPE", None)
        if sigpipe:  # ie: POSIX
            signal.signal(sigpipe, self.handle_signal)
        self.opacity = 100
        self.pct = 0
        self.start_time = monotonic()
        self.had_top_level_focus = False
        self.fd_watch = 0
        self.connect("notify::has-toplevel-focus", self._focus_change)

    def cancel_io_watch(self) -> None:
        fd = self.fd_watch
        if fd:
            self.fd_watch = 0
            GLib.source_remove(fd)

    def run(self) -> ExitValue:
        events = GLib.IO_IN | GLib.IO_HUP | GLib.IO_ERR
        self.fd_watch = GLib.io_add_watch(sys.stdin.fileno(), GLib.PRIORITY_LOW, events, self.read_stdin)
        self.show_all()
        force_focus()
        self.present()
        self.timeout_timer = GLib.timeout_add(TIMEOUT * 1000, self.timeout)
        self.pulse_timer = GLib.timeout_add(100, self.pulse)
        scrash = envint("XPRA_SPLASH_CRASH", -1)
        if scrash >= 0:
            def crash():
                import ctypes  # pylint: disable=import-outside-toplevel
                ctypes.string_at(0)

            GLib.timeout_add(scrash, crash)
        self.start_time = monotonic()
        Gtk.main()
        return self.exit_code or 0

    def _focus_change(self, *args) -> None:
        if WIN32 and monotonic() - self.start_time < 1:
            # ignore initial focus events on win32
            return
        has = self.has_toplevel_focus()
        had = self.had_top_level_focus
        log(f"_focus_change{args} {had=}, {has=}")
        if had and not has:
            if FOCUS_EXIT:
                self.exit()
        elif has:
            self.had_top_level_focus = True

    def timeout(self) -> None:
        log("timeout()")
        self.timeout_timer = 0
        self.exit_code = ExitCode.TIMEOUT
        self.show_progress_value(100)
        self.progress_bar.set_text("timeout")
        self.progress_bar.set_show_text(True)
        self.exit()

    def cancel_timeout_timer(self) -> None:
        tt = self.timeout_timer
        if tt:
            self.timeout_timer = 0
            GLib.source_remove(tt)

    def pulse(self) -> bool:
        if not self.current_label_text:
            return True
        if self.pct < 100:
            pulse_char = PULSE_CHARS[self.pulse_counter % len(PULSE_CHARS)]
        else:
            pulse_char = DONE_CHAR
        self.labels[-1].set_label(f"  {pulse_char} {self.current_label_text}")
        self.pulse_counter += 1
        return True

    def read_stdin(self, fd, cb_condition) -> bool:
        log(f"read_stdin({fd}, {cb_condition})")
        if cb_condition in (GLib.IO_HUP, GLib.IO_ERR):
            self.fd_watch = 0
            self.exit_code = ExitCode.CONNECTION_LOST
            self.exit()
            return False
        if cb_condition == GLib.IO_IN:
            try:
                line = sys.stdin.readline()
                GLib.idle_add(self.handle_stdin_line, line)
                sleep(READ_SLEEP)
                return True
            except OSError:
                self.fd_watch = 0
                self.exit_code = ExitCode.FAILURE
                self.exit()
                return False
        # unexpected condition - continue
        return True

    def handle_stdin_line(self, line: str) -> None:
        parts = line.rstrip("\n\r").split(":", 1)
        log("handle_stdin_line(%r)", line)
        pct = self.pct
        if parts[0]:
            try:
                pct = int(parts[0])
            except ValueError:
                pass
            else:
                if pct > 0:
                    self.show_progress_value(pct)
        if len(parts) >= 2:
            text = parts[1]
            if pct == 0:
                self.set_title(text)
                self.title_label.set_text(text)
                self.current_label_text = ""
            else:
                self.current_label_text = text
                if self.pct != pct:
                    for i in range(len(self.labels) - 1):
                        next_line = self.labels[i + 1].get_label()
                        for c in PULSE_CHARS:
                            next_line = next_line.replace(c, DONE_CHAR)
                        self.labels[i].set_label(next_line)
            self.pulse_counter = 0
        self.pct = pct
        if pct > 0:
            self.pulse()

    def show_progress_value(self, pct: int) -> None:
        self.cancel_progress_timer()
        GLib.idle_add(self.progress_bar.set_fraction, pct / 100.0)
        if pct >= 100:
            self.cancel_pulse_timer()
            self.cancel_fade_out_timer()
            self.fade_out_timer = GLib.timeout_add(SPLASH_EXIT_DELAY * 1000 // 100, self.fade_out)
            self.cancel_exit_timer()

            def exit_splash():
                self.exit_timer = 0
                self.exit()

            self.exit_timer = GLib.timeout_add(SPLASH_EXIT_DELAY * 1000, exit_splash)
        else:
            self.progress_timer = GLib.timeout_add(40, self.increase_fraction, pct)
            self.opacity = min(100, max(50, 130 - pct))
            self.set_opacity(self.opacity / 100.0)
        set_window_progress(self, pct)

    def cancel_exit_timer(self) -> None:
        et = self.exit_timer
        if et:
            self.exit_timer = 0
            GLib.source_remove(et)

    def cancel_fade_out_timer(self) -> None:
        fot = self.fade_out_timer
        if fot:
            self.fade_out_timer = 0
            GLib.source_remove(fot)

    def cancel_progress_timer(self) -> None:
        pt = self.progress_timer
        if pt:
            self.progress_timer = 0
            GLib.source_remove(pt)

    def cancel_pulse_timer(self) -> None:
        pt = self.pulse_timer
        if pt:
            self.pulse_timer = 0
            GLib.source_remove(pt)

    def increase_fraction(self, pct: int, inc=1, max_increase=10) -> bool:
        log("increase_fraction%s", (pct, inc, max_increase))
        self.cancel_progress_timer()
        GLib.idle_add(self.progress_bar.set_fraction, (pct + inc) / 100.0)
        if inc < max_increase:
            self.progress_timer = GLib.timeout_add(40 + 20 * (2 + inc) ** 2, self.increase_fraction, pct, inc + 1,
                                                   max_increase)
        return False

    def fade_out(self) -> bool:
        self.opacity = max(0, self.opacity - 1)
        actual = int(ignorewarnings(self.get_opacity) * 100)
        if actual > self.opacity:
            self.set_opacity(self.opacity / 100.0)
        if actual <= 0:
            self.fade_out_timer = 0
        return actual > 0

    def exit(self, *args) -> None:
        log("exit%s calling %s", args, Gtk.main_quit)
        if self.exit_code is None:
            self.exit_code = 0
        self.cancel_io_watch()
        self.cancel_progress_timer()
        self.cancel_timeout_timer()
        self.cancel_fade_out_timer()
        self.cancel_pulse_timer()
        self.cancel_exit_timer()
        Gtk.main_quit()

    def handle_signal(self, signum, frame=None) -> None:
        log("handle_signal(%s, %s)", SIGNAMES.get(signum, signum), frame)
        self.exit_code = 128 - (signum or 0)
        GLib.idle_add(self.exit)


def main(_args) -> ExitValue:
    import os
    if os.environ.get("XPRA_HIDE_DOCK") is None:
        os.environ["XPRA_HIDE_DOCK"] = "1"
    from xpra.platform import program_context
    with program_context("splash", "Splash"):
        Gtk.Window.set_auto_startup_notification(setting=False)
        w = SplashScreen()
        from xpra.gtk.window import add_close_accel
        add_close_accel(w, Gtk.main_quit)
        return w.run()


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os
import socket
import platform
from typing import Any, Final

import xpra
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.os_util import BITS, POSIX, WIN32, OSX
from xpra.util.io import get_util_logger
from xpra.util.system import get_linux_distribution, platform_release, platform_name
from xpra.common import FULL_INFO

XPRA_VERSION: Final[str] = xpra.__version__
XPRA_NUMERIC_VERSION: tuple[int] = xpra.__version_info__

CHECK_SSL: bool = envbool("XPRA_VERSION_CHECK_SSL", True)
SSL_CAFILE: str = ""
if WIN32:
    try:
        import certifi

        SSL_CAFILE = certifi.where()
    except (ImportError, AttributeError):
        get_util_logger().error("failed to locate SSL ca file", exc_info=True)
SSL_CAFILE = os.environ.get("XPRA_SSL_CAFILE", SSL_CAFILE)


def log(msg, *args, **kwargs) -> None:
    get_util_logger().debug(msg, *args, **kwargs)


def warn(msg, *args, **kwargs) -> None:
    get_util_logger().warn(msg, *args, **kwargs)


def vparts(vstr: str, n=1) -> str:
    return ".".join(vstr.split(".")[:n])


def version_str() -> str:
    rstr = revision_str()
    return XPRA_VERSION if not rstr else XPRA_VERSION + "-" + rstr


def full_version_str() -> str:
    rstr = version_str()
    try:
        # pylint: disable=import-outside-toplevel
        from xpra.src_info import BRANCH
    except ImportError:
        pass
    else:
        if BRANCH == "master":
            rstr += " beta"
    try:
        from xpra.build_info import build
        btype = build["type"]
        if btype == "light":
            rstr += " (light build)"
    except (ImportError, KeyError):
        pass
    return rstr


def caps_to_version(caps: typedict) -> str:
    return caps.strget("version", "0") + "-" + caps_to_revision(caps)


def caps_to_revision(caps: typedict) -> str:
    revision = caps.intget("revision")
    local_modifications = caps.intget("local_modifications")
    commit = caps.strget("commit")
    branch = caps.strget("branch")
    return make_revision_str(revision, local_modifications, branch, commit)


def revision_str() -> str:
    try:
        # pylint: disable=import-outside-toplevel
        from xpra.src_info import REVISION, LOCAL_MODIFICATIONS, BRANCH, COMMIT
    except ImportError:
        pass
    else:
        return make_revision_str(REVISION, LOCAL_MODIFICATIONS, BRANCH, COMMIT)
    return ""


def make_revision_str(revision, local_modifications, branch, commit) -> str:
    rstr = ""
    try:
        if isinstance(revision, int):
            rstr += f"r{revision}"
        if isinstance(local_modifications, int) and local_modifications > 0:
            rstr += "M"
        if branch == "master" and commit:
            rstr += f" ({commit})"
    except TypeError:
        get_util_logger().debug("make_revision_str%s", (revision, local_modifications, branch, commit), exc_info=True)
    return rstr


def version_compat_check(remote_version) -> str | None:
    if not remote_version:
        msg = "remote version is not available"
        log(msg)
        return msg
    try:
        rv = parse_version(remote_version)
    except ValueError:
        warn(f"Warning: failed to parse remote version {remote_version!r}")
        return None
    rvstr = ".".join(str(part) for part in rv)
    if rv == XPRA_NUMERIC_VERSION:
        log(f"identical remote version: {rvstr!r}")
        return None
    if rv[:2] == XPRA_NUMERIC_VERSION[:2]:
        log(f"identical major.minor in remote version: {rvstr!r}")
        return None
    try:
        if rv[0:2] < (3, 0):
            # this is the oldest version we support
            msg = f"remote version {rvstr!r} is too old, sorry"
            log(msg)
            return msg
        if rv[0:2] < (3, 1, 9):
            warn(f"Warning: remote version {rvstr!r} is outdated and buggy")
            return None
    except TypeError as e:
        msg = f"invalid remote version {rvstr!r}: {e}"
        log(msg)
        return msg
    if rv[0] > 0:
        log(f"newer remote version {rvstr!r} should work, we'll see..")
        return None
    log(f"local version {XPRA_VERSION!r} should be compatible with remote version {rvstr!r}")
    return None


def get_host_info(full_info: int = 1) -> dict[str, Any]:
    # this function is for non UI thread info
    info: dict[str, Any] = {}
    if full_info > 1:
        info |= {
            "byteorder": sys.byteorder,
            "python": {
                "bits": BITS,
                "full_version": sys.version,
                "version": ".".join(str(x) for x in sys.version_info[:3]),
            },
        }
    if full_info > 0:
        try:
            hostname = socket.gethostname()
            if hostname:
                info["hostname"] = hostname
        except OSError:
            pass
        if POSIX:
            info |= {
                "uid": os.getuid(),
                "gid": os.getgid(),
            }
    return info


def get_version_info(full: int = 1) -> dict[str, Any]:
    info: dict[str, Any] = {"version": vparts(XPRA_VERSION, full + 1)}
    if full > 0:
        try:
            # pylint: disable=import-outside-toplevel
            from xpra.src_info import LOCAL_MODIFICATIONS, REVISION, COMMIT, BRANCH
            for k, v in {
                "local_modifications": LOCAL_MODIFICATIONS,
                "revision": REVISION,
                "branch": BRANCH,
                "commit": COMMIT,
            }.items():
                if v is not None and v != "unknown":
                    info[k] = v
        except ImportError as e:
            warn("missing some source information: %s", e)
        info.update(get_build_info(full))
    return info


def get_build_info(full: int = 1) -> dict[str, Any]:
    info: dict[str, Any] = {}

    try:
        from xpra import build_info  # pylint: disable=import-outside-toplevel
    except ImportError:
        return {}

    build = getattr(build_info, "build", {})

    def add_attrs(attrs: dict[str, str]) -> None:
        for k, type_info in attrs.items():
            v = build.get(k, None)
            if v is not None:
                if type_info.endswith("_VERSION"):
                    v = parse_version(v)
                info[k] = v

    try:
        add_attrs({
            "date": "BUILD_DATE",
            "time": "BUILD_TIME",
        })
        if full > 0:
            add_attrs({
                "bit": "BUILD_BIT",
                "cpu": "BUILD_CPU",
                "type": "BUILD_TYPE",
                "compiler": "COMPILER_VERSION",
                "nvcc": "NVCC_VERSION",
                "linker": "LINKER_VERSION",
                "python": "PYTHON_VERSION",
                "cython": "CYTHON_VERSION",
            })
        if full > 1:
            # record library versions:
            info["lib"] = {k.lstrip("lib_"): parse_version(getattr(build_info, k))
                           for k in dir(build_info) if k.startswith("lib_")}
    except Exception as e:
        warn("missing some build information: %s", e)
    log(f"get_build_info({full})={info}")
    return info


def parse_version(v) -> tuple[Any]:
    if isinstance(v, str) and v:
        def maybeint(value: str) -> int | str:
            try:
                return int(value)
            except ValueError:
                return value

        v = tuple(maybeint(x) for x in v.split("-")[0].split("."))
    return tuple(v or ())


def vtrim(v, parts=FULL_INFO + 1):
    if isinstance(v, (list, tuple)):
        return v[:parts]
    return v


def dict_version_trim(d, parts=FULL_INFO + 1) -> dict:
    """
    trims version numbers from info dictionaries
    """

    def vfilt(k, v):
        if k.endswith("version") and isinstance(v, (list, tuple)):
            v = vtrim(v, parts)
        elif isinstance(v, dict):
            return k, dict_version_trim(v, parts)
        return k, v

    return dict(vfilt(k, v) for k, v in d.items())


def do_get_platform_info() -> dict[str, Any]:
    info: dict[str, Any] = {}
    if POSIX and not OSX:
        ld = get_linux_distribution()
        ldvalid = tuple(x for x in ld if x not in ("", "unknown", "n/a"))
        if ldvalid:
            info["linux_distribution"] = ld
    try:
        release = platform_release(platform.release())
    except OSError:
        log("do_get_platform_info()", exc_info=True)
        release = "unknown"
    info |= {
        "": sys.platform,
        "name": platform_name(sys.platform, info.get("linux_distribution") or release),
        "release": platform.release(),
        "sysrelease": release,
        "platform": platform.platform(),
        "machine": platform.machine(),
        "architecture": platform.architecture(),
    }
    try:
        from xpra.util.system import get_processor_name
        info["processor"] = get_processor_name()
    except Exception:
        log("failed to query processor", exc_info=True)
    return info


# cache the output:
platform_info_cache = None


def get_platform_info() -> dict[str, Any]:
    global platform_info_cache
    if platform_info_cache is None:
        platform_info_cache = do_get_platform_info()
    return platform_info_cache


def get_version_from_url(url: str) -> tuple[int, ...]:
    try:
        import ssl
        from urllib.request import urlopen
        from urllib.error import HTTPError
    except ImportError as e:
        log("get_version_from_url(%s) urllib2 not found: %s", url, e)
        return ()
    try:
        context = ssl.create_default_context(cafile=SSL_CAFILE)
        response = urlopen(url, context=context)
        latest_version = response.read().rstrip(b"\n\r")
        latest_version_no = tuple(int(y) for y in latest_version.split(b"."))
        log("get_version_from_url(%s)=%s", url, latest_version_no)
        return latest_version_no
    except HTTPError as e:
        log(f"get_version_from_url({url!r}) {e}", exc_info=e.code != 404)
    except Exception:
        get_util_logger().error(f"Error accessing {url!r}", exc_info=True)
    return ()


PLATFORM_FRIENDLY_NAMES: dict[str, str] = {
    "linux2": "LINUX",
    "win": "WINDOWS",
    "darwin": "OSX",
}


def get_branch() -> str:
    BRANCH = os.environ.get("XPRA_CURRENT_VERSION_BRANCH", "")
    if BRANCH:
        return BRANCH
    try:
        from xpra.src_info import BRANCH
        return BRANCH
    except ImportError as e:
        log(f"unknown branch: {e}")
    return ""


def get_latest_version() -> bool | None | tuple[int, ...]:
    CURRENT_VERSION_URL = ("https" if CHECK_SSL else "http") + "://xpra.org/CURRENT_VERSION"
    BRANCH = get_branch()
    branch_strs = []
    if BRANCH not in ("", "master"):
        branch_parts = BRANCH.split(".")            # ie: "v6.2.x" -> ["v6", "2", "x"]
        if branch_parts[-1] == "x":
            branch_parts = branch_parts[:-1]        # ie: ["v6", "2"]
        branch_strs.append("_" + "_".join(branch_parts[:2]))    # ie: "_v6_2"
        if len(branch_parts) >= 2:
            branch_strs.append("_" + branch_parts[0])           # ie: "_v6"
    branch_strs.append("")
    platname = PLATFORM_FRIENDLY_NAMES.get(sys.platform, sys.platform)
    arch = get_platform_info().get("machine")
    for branch_str in branch_strs:
        for url in (
            f"{CURRENT_VERSION_URL}{branch_str}_{platname}_{arch}?{XPRA_VERSION}",
            f"{CURRENT_VERSION_URL}{branch_str}_{platname}?{XPRA_VERSION}",
            f"{CURRENT_VERSION_URL}{branch_str}?{XPRA_VERSION}",
        ):
            latest_version_no = get_version_from_url(url)
            if latest_version_no:
                return latest_version_no
    return ()


def version_update_check() -> bool | None | tuple[int, ...]:
    FAKE_NEW_VERSION = envbool("XPRA_FAKE_NEW_VERSION", False)
    latest_version_no = get_latest_version()
    if not latest_version_no:
        log("version_update_check() failed to contact version server")
        return None
    if latest_version_no > XPRA_NUMERIC_VERSION or FAKE_NEW_VERSION:
        log("version_update_check() newer version found")
        log(f" local version is {XPRA_NUMERIC_VERSION} and the latest version available is {latest_version_no}")
        return latest_version_no
    return False



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os.path

from xpra.os_util import WIN32, OSX
from xpra.log import Logger, consume_verbose_argv

log = Logger("audio")

default_icon_path = None


def set_icon_path(v) -> None:
    global default_icon_path
    default_icon_path = v


def add_audio_tagging_env(env_dict: dict = os.environ, icon_path: str = "") -> None:
    """
        This is called audio-tagging in PulseAudio, see:
        http://pulseaudio.org/wiki/ApplicationProperties
        http://0pointer.de/blog/projects/tagging-audio.html
    """
    from xpra.util.version import XPRA_VERSION
    env_dict |= {
        "PULSE_PROP_application.name": "xpra",
        "PULSE_PROP_application.id": "xpra",
        "PULSE_PROP_application.version": XPRA_VERSION,
        "PULSE_PROP_media.role": "music",
    }
    if not icon_path:
        icon_path = default_icon_path
    if icon_path and os.path.exists(icon_path):
        env_dict["PULSE_PROP_application.icon_name"] = str(icon_path)


try:
    # use "none" on win32 and osx:
    if WIN32 or OSX:
        from xpra.audio.pulseaudio import none_impl as _pulseaudio_util
    else:
        from xpra.audio.pulseaudio import pactl_impl as _pulseaudio_util
except ImportError as e:
    # fallback forks a process and parses the output:
    log("cannot import default pulseaudio util: %s", e)
    log("using pulseaudio none fallback")
    del e
    from xpra.audio.pulseaudio import none_impl as _pulseaudio_util

get_info = _pulseaudio_util.get_info
has_pa = _pulseaudio_util.has_pa
get_pa_device_options = _pulseaudio_util.get_pa_device_options
get_default_sink = _pulseaudio_util.get_default_sink
get_pulse_server = _pulseaudio_util.get_pulse_server
get_pulse_id = _pulseaudio_util.get_pulse_id
get_pulse_cookie_hash = _pulseaudio_util.get_pulse_cookie_hash
get_pactl_server = _pulseaudio_util.get_pactl_server
set_source_mute = _pulseaudio_util.set_source_mute
set_sink_mute = _pulseaudio_util.set_sink_mute


def main():
    from xpra.platform import program_context
    from xpra.log import enable_color
    from xpra.util.str_fn import print_nested_dict
    with program_context("Pulseaudio-Info"):
        enable_color()
        consume_verbose_argv(sys.argv, "audio")
        i = get_info()
        print_nested_dict(i)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import struct
import socket
import yaml

FLAGS_YAML = 0x4

encoding = "h264"
encoder = "openh264"
width = 640
height = 480
pixel_format = "YUV420P"


def main(args):
    sockpath = args[1] if len(args) > 1 else "/run/xpra/encoder/socket"
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
        sock.settimeout(None)
        sock.connect(sockpath)

        def send(*args) -> None:
            payload = yaml.dump(list(args)).encode("utf-8")
            header = struct.pack(b"!cBBBL", b"P", FLAGS_YAML, 0, 0, len(payload))
            sock.send(header + payload)

        bufs = []

        def parse() -> tuple:
            if not bufs:
                return ()
            data = b"".join(bufs)
            bufs[:] = [data]
            if len(data) <= 8:
                return ()
            _, _, _, index, size = struct.unpack(b"!cBBBL", data[:8])
            assert index == 0
            if len(data) - 8 < size:
                return ()
            payload = data[8:8 + size]
            bufs[:] = [data[8 + size:]]
            return yaml.load(payload, Loader=yaml.SafeLoader)

        def recv() -> tuple:
            while True:
                packet = parse()
                if packet:
                    return packet
                data = sock.recv(1024)
                if not data:
                    return ()
                bufs.append(data)

        send("hello", {"chunks": False, "yaml": True, "version": "6.3", "encoding": {"core": ["rgb32", ]}})
        packet = recv()
        assert packet[0] == "hello"
        packet = recv()
        assert packet[0] == "encodings", f"got {packet[0]!r}"
        packet = recv()
        assert packet[0] == "startup-complete", f"got {packet[0]!r}"
        send("context-request", 0, encoder, encoding, width, height, pixel_format, {})
        packet = recv()
        assert packet[0] == "context-response", f"got {packet[0]!r}"
        metadata = {
            "x": 0,
            "y": 0,
            "width": width,
            "height": height,
            "pixel_format": pixel_format,
            "depth": 24,
            "rowstride": [width, width//2, width//2],
            "planes": 3,
        }
        pixels = [
            b"P"*(width * height),
            b"Q"*(width // 2 * height // 2),
            b"R"*(width // 2 * height // 2),
        ]
        send("context-compress", 0, metadata, pixels, {})
        packet = recv()
        assert packet[0] == "context-data", f"got {packet[0]!r}"
        data = packet[2]
        with open(f"./frame.{encoding}", "wb") as f:
            f.write(data)
        sock.close()
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))
